# coding: utf-8
# 2021/6/21 @ tongshiwei
from tqdm import tqdm
from longling import as_list
from copy import deepcopy
from .utils import POrderedDict

__all__ = ["ranking_report"]


def ranking_auc(ranked_label):
    """
    Examples
    --------
    >>> ranking_auc([1, 1, 0, 0, 0])
    1.0
    >>> ranking_auc([0, 1, 0, 1, 0])
    0.5
    >>> ranking_auc([1, 0, 1, 0, 0])
    0.8333333333333334
    >>> ranking_auc([0, 0, 0, 1, 1])
    0.0
    """
    pos_num = sum(ranked_label)
    neg_num = len(ranked_label) - sum(ranked_label)
    if pos_num * neg_num == 0:  # pragma: no cover
        return 1
    return sum(
        [len(ranked_label[i + 1:]) - sum(ranked_label[i + 1:]) for i, score in enumerate(ranked_label) if score == 1]
    ) / (pos_num * neg_num)


def ranking_report(y_true, y_pred, k: (int, list) = None, continuous=False, coerce="ignore", pad_pred=-100,
                   metrics=None, bottom=False, verbose=True) -> POrderedDict:
    r"""

    Parameters
    ----------
    y_true
    y_pred
    k
    continuous
    coerce
    pad_pred
    metrics
    bottom
    verbose

    Returns
    -------

    Examples
    --------
    >>> y_true = [[1, 0, 0], [0, 0, 1]]
    >>> y_pred = [[0.75, 0.5, 1], [1, 0.2, 0.1]]
    >>> ranking_report(y_true, y_pred)  # doctest: +NORMALIZE_WHITESPACE
           ndcg@k  precision@k  recall@k  f1@k  len@k  support@k
    1   1.000000     0.000000       0.0   0.0    1.0          2
    3   0.565465     0.333333       1.0   0.5    3.0          2
    5   0.565465     0.333333       1.0   0.5    3.0          2
    10  0.565465     0.333333       1.0   0.5    3.0          2
    auc: 0.250000	map: 0.416667	mrr: 0.416667	coverage_error: 2.500000	ranking_loss: 0.750000	len: 3.000000
    support: 2
    >>> ranking_report(y_true, y_pred, k=[1, 3, 5])  # doctest: +NORMALIZE_WHITESPACE
           ndcg@k  precision@k  recall@k  f1@k  len@k  support@k
    1   1.000000     0.000000       0.0   0.0    1.0          2
    3   0.565465     0.333333       1.0   0.5    3.0          2
    5   0.565465     0.333333       1.0   0.5    3.0          2
    auc: 0.250000	map: 0.416667	mrr: 0.416667	coverage_error: 2.500000	ranking_loss: 0.750000	len: 3.000000
    support: 2
    >>> ranking_report(y_true, y_pred, bottom=True)  # doctest: +NORMALIZE_WHITESPACE
              ndcg@k  precision@k  recall@k  f1@k  len@k  support@k  ndcg@k(B) \
    1   1.000000     0.000000       0.0   0.0    1.0          2   1.000000
    3   0.565465     0.333333       1.0   0.5    3.0          2   0.806574
    5   0.565465     0.333333       1.0   0.5    3.0          2   0.806574
    10  0.565465     0.333333       1.0   0.5    3.0          2   0.806574
    <BLANKLINE>
        precision@k(B)  recall@k(B)   f1@k(B)  len@k(B)  support@k(B)
    1         0.500000         0.25  0.333333       1.0             2
    3         0.666667         1.00  0.800000       3.0             2
    5         0.666667         1.00  0.800000       3.0             2
    10        0.666667         1.00  0.800000       3.0             2
    auc: 0.250000	map: 0.416667	mrr: 0.416667	coverage_error: 2.500000	ranking_loss: 0.750000	len: 3.000000
    support: 2	map(B): 0.708333	mrr(B): 0.750000
    >>> ranking_report(y_true, y_pred, bottom=True, metrics=["auc"])  # doctest: +NORMALIZE_WHITESPACE
    auc: 0.250000   len: 3.000000	support: 2
    >>> y_true = [[0.9, 0.7, 0.1], [0, 0.5, 1]]
    >>> y_pred = [[0.75, 0.5, 1], [1, 0.2, 0.1]]
    >>> ranking_report(y_true, y_pred, continuous=True)  # doctest: +NORMALIZE_WHITESPACE
          ndcg@k  len@k  support@k
    3   0.675647    3.0          2
    5   0.675647    3.0          2
    10  0.675647    3.0          2
    mrr: 0.750000	len: 3.000000	support: 2
    >>> y_true = [[1, 0], [0, 0, 1]]
    >>> y_pred = [[0.75, 0.5], [1, 0.2, 0.1]]
    >>> ranking_report(y_true, y_pred)  # doctest: +NORMALIZE_WHITESPACE
        ndcg@k  precision@k  recall@k      f1@k  len@k  support@k
    1     1.00     0.500000       0.5  0.500000    1.0          2
    3     0.75     0.416667       1.0  0.583333    2.5          2
    5     0.75     0.416667       1.0  0.583333    2.5          2
    10    0.75     0.416667       1.0  0.583333    2.5          2
    auc: 0.500000	map: 0.666667	mrr: 0.666667	coverage_error: 2.000000	ranking_loss: 0.500000	len: 2.500000
    support: 2
    >>> ranking_report(y_true, y_pred, coerce="abandon")  # doctest: +NORMALIZE_WHITESPACE
       ndcg@k  precision@k  recall@k  f1@k  len@k  support@k
    1     1.0     0.500000       0.5   0.5    1.0          2
    3     0.5     0.333333       1.0   0.5    3.0          1
    auc: 0.500000	map: 0.666667	mrr: 0.666667	coverage_error: 2.000000	ranking_loss: 0.500000	len: 2.500000
    support: 2
    >>> ranking_report(y_true, y_pred, coerce="padding")  # doctest: +NORMALIZE_WHITESPACE
        ndcg@k  precision@k  recall@k      f1@k  len@k  support@k
    1     1.00     0.500000       0.5  0.500000    1.0          2
    3     0.75     0.416667       1.0  0.583333    2.5          2
    5     0.75     0.416667       1.0  0.583333    2.5          2
    10    0.75     0.416667       1.0  0.583333    2.5          2
    auc: 0.500000	map: 0.666667	mrr: 0.666667	coverage_error: 2.000000	ranking_loss: 0.500000	len: 2.500000
    support: 2
    >>> ranking_report(y_true, y_pred, bottom=True)  # doctest: +NORMALIZE_WHITESPACE
        ndcg@k  precision@k  recall@k      f1@k  len@k  support@k  ndcg@k(B)  \
    1     1.00     0.500000       0.5  0.500000    1.0          2   1.000000
    3     0.75     0.416667       1.0  0.583333    2.5          2   0.846713
    5     0.75     0.416667       1.0  0.583333    2.5          2   0.846713
    10    0.75     0.416667       1.0  0.583333    2.5          2   0.846713
    <BLANKLINE>
        precision@k(B)  recall@k(B)   f1@k(B)  len@k(B)  support@k(B)
    1         0.500000          0.5  0.500000       1.0             2
    3         0.583333          1.0  0.733333       2.5             2
    5         0.583333          1.0  0.733333       2.5             2
    10        0.583333          1.0  0.733333       2.5             2
    auc: 0.500000	map: 0.666667	mrr: 0.666667	coverage_error: 2.000000	ranking_loss: 0.500000	len: 2.500000
    support: 2	map(B): 0.791667	mrr(B): 0.750000
    >>> ranking_report(y_true, y_pred, bottom=True, coerce="abandon")  # doctest: +NORMALIZE_WHITESPACE
       ndcg@k  precision@k  recall@k  f1@k  len@k  support@k  ndcg@k(B)  \
    1     1.0     0.500000       0.5   0.5    1.0          2   1.000000
    3     0.5     0.333333       1.0   0.5    3.0          1   0.693426
    <BLANKLINE>
       precision@k(B)  recall@k(B)  f1@k(B)  len@k(B)  support@k(B)
    1        0.500000          0.5      0.5       1.0             2
    3        0.666667          1.0      0.8       3.0             1
    auc: 0.500000	map: 0.666667	mrr: 0.666667	coverage_error: 2.000000	ranking_loss: 0.500000
    len: 2.500000	support: 2	map(B): 0.791667	mrr(B): 0.750000
    >>> ranking_report(y_true, y_pred, bottom=True, coerce="padding")  # doctest: +NORMALIZE_WHITESPACE
        ndcg@k  precision@k  recall@k      f1@k  len@k  support@k  ndcg@k(B)  \
    1     1.00     0.500000       0.5  0.500000    1.0          2   1.000000
    3     0.75     0.416667       1.0  0.583333    2.5          2   0.846713
    5     0.75     0.416667       1.0  0.583333    2.5          2   0.846713
    10    0.75     0.416667       1.0  0.583333    2.5          2   0.846713
    <BLANKLINE>
        precision@k(B)  recall@k(B)   f1@k(B)  len@k(B)  support@k(B)
    1             0.50          0.5  0.500000       1.0             2
    3             0.50          1.0  0.650000       3.0             2
    5             0.30          1.0  0.452381       5.0             2
    10            0.15          1.0  0.257576      10.0             2
    auc: 0.500000	map: 0.666667	mrr: 0.666667	coverage_error: 2.000000	ranking_loss: 0.500000	len: 2.500000
    support: 2	map(B): 0.791667	mrr(B): 0.750000
    """
    import numpy as np
    from collections import OrderedDict
    from sklearn.metrics import (
        label_ranking_average_precision_score,
        ndcg_score, label_ranking_loss, coverage_error
    )
    assert coerce in {"ignore", "abandon", "raise", "padding"}
    if metrics is None:
        metrics = ["mrr", "ndcg"]
        if continuous is False:
            metrics.extend(["auc", "map", "coverage_error", "ranking_loss", "precision", "recall", "f1"])
    metrics = set(metrics)

    if k is not None:
        k = as_list(k)
    else:
        if continuous is True:
            k = [3, 5, 10]
        else:
            k = [1, 3, 5, 10]

    results = {
        "auc": [],
        "map": [],
        "mrr": [],
        "coverage_error": [],
        "ranking_loss": [],
        "len": [],
        "support": [],
    }
    if bottom:
        results.update({
            "map(B)": [],
            "mrr(B)": [],
        })
    k_results = {}
    for _k in k:
        k_results[_k] = {
            "ndcg@k": [],
            "precision@k": [],
            "recall@k": [],
            "f1@k": [],
            "len@k": [],
            "support@k": [],
        }
        if bottom:
            k_results[_k].update({
                "ndcg@k(B)": [],
                "precision@k(B)": [],
                "recall@k(B)": [],
                "f1@k(B)": [],
                "len@k(B)": [],
                "support@k(B)": [],
            })
    suffix = [""]
    if bottom:
        suffix += ["(B)"]

    for label, pred in tqdm(zip(y_true, y_pred), "ranking metrics", disable=not verbose):
        if continuous is False and "map" in metrics:
            results["map"].append(label_ranking_average_precision_score([label], [pred]))
            if bottom:
                results["map(B)"].append(label_ranking_average_precision_score(
                    [(1 - np.asarray(label)).tolist()],
                    [(-np.asarray(pred)).tolist()]
                ))

        if len(label) > 1 and continuous is False:
            if "coverage_error" in metrics:
                results["coverage_error"].append(coverage_error([label], [pred]))
            if "ranking_loss" in metrics:
                results["ranking_loss"].append(label_ranking_loss([label], [pred]))

        results["len"].append(len(label))
        results["support"].append(1)
        label_pred = list(sorted(zip(label, pred), key=lambda x: x[1], reverse=True))
        sorted_label = list(zip(*label_pred))[0]
        if "auc" in metrics:
            results["auc"].append(ranking_auc(sorted_label))
        if "mrr" in metrics:
            try:
                results["mrr"].append(1 / (np.asarray(sorted_label).nonzero()[0][0] + 1))
            except IndexError:  # pragma: no cover
                pass
            try:
                if bottom:
                    results["mrr(B)"].append(1 / (np.asarray(sorted_label[::-1]).nonzero()[0][0] + 1))
            except IndexError:  # pragma: no cover
                pass

        if metrics & {"ndcg", "precision", "recall", "f1"}:
            for _k in k:
                for _suffix in suffix:
                    if _suffix == "":
                        _label_pred = deepcopy(label_pred)
                        if len(_label_pred) < _k:
                            if coerce == "ignore":
                                pass
                            elif coerce == "abandon":
                                continue
                            elif coerce == "raise":
                                raise ValueError("Not enough value: %s vs target %s" % (len(_label_pred), _k))
                            elif coerce == "padding":  # pragma: no cover
                                _label_pred += [(0, pad_pred)] * (_k - len(_label_pred))
                        k_label_pred = label_pred[:_k]
                        total_label = sum(label)
                    else:
                        inv_label_pred = [(1 - _l, -p) for _l, p in label_pred][::-1]
                        if len(inv_label_pred) < _k:
                            if coerce == "ignore":
                                pass
                            elif coerce == "abandon":
                                continue
                            elif coerce == "raise":  # pragma: no cover
                                raise ValueError("Not enough value: %s vs target %s" % (len(inv_label_pred), _k))
                            elif coerce == "padding":
                                inv_label_pred += [(0, pad_pred)] * (_k - len(inv_label_pred))
                        k_label_pred = inv_label_pred[:_k]
                        total_label = len(label) - sum(label)

                    if not k_label_pred:  # pragma: no cover
                        continue
                    k_label, k_pred = list(zip(*k_label_pred))
                    if "ndcg" in metrics:
                        if len(k_label) == 1 and "ndcg" in metrics:
                            k_results[_k]["ndcg@k%s" % _suffix].append(1)
                        else:
                            k_results[_k]["ndcg@k%s" % _suffix].append(ndcg_score([k_label], [k_pred]))
                    p = sum(k_label) / len(k_label)
                    r = sum(k_label) / total_label if total_label else 0
                    if "precision" in metrics:
                        k_results[_k]["precision@k%s" % _suffix].append(p)
                    if "recall" in metrics:
                        k_results[_k]["recall@k%s" % _suffix].append(r)
                    if "f1" in metrics:
                        k_results[_k]["f1@k%s" % _suffix].append(2 * p * r / (p + r) if p + r else 0)
                    k_results[_k]["len@k%s" % _suffix].append(len(k_label))
                    k_results[_k]["support@k%s" % _suffix].append(1)

    ret = POrderedDict()
    for key, value in results.items():
        if value:
            if key == "support":
                ret[key] = np.sum(value).item()
            else:
                ret[key] = np.mean(value).item()

    if metrics & {"ndcg", "precision", "recall", "f1"}:
        for k, key_value in k_results.items():
            ret[k] = OrderedDict()
            for key, value in key_value.items():
                if value:
                    if key in {"support@k", "support@k(B)"}:
                        ret[k][key] = np.sum(value).item()
                    else:
                        ret[k][key] = np.mean(value).item()
    return ret



# coding: utf-8
# create by tongshiwei on 2020-11-21

import pandas as pd

__all__ = ["ID2Feature"]


class ID2Feature(object):
    """
    Examples
    --------
    >>> import pandas as pd
    >>> df = pd.DataFrame({"id": [0, 1, 2, 3, 4], "numeric": [1, 2, 3, 4, 5], "text": ["a", "b", "c", "d", "e"]})
    >>> i2f = ID2Feature(df, id_field="id", set_index=True)
    >>> i2f[2]
    numeric    3
    text       c
    Name: 2, dtype: object
    >>> i2f[[2, 3]]["numeric"]
    id
    2    3
    3    4
    Name: numeric, dtype: int64
    >>> i2f(2)
    [3, 'c']
    >>> i2f([2, 3])
    [[3, 'c'], [4, 'd']]
    """

    def __init__(self, feature_df: pd.DataFrame, id_field=None, set_index=False):
        self.feature_df = feature_df
        if set_index:
            self.feature_df.set_index(id_field, inplace=True)

    def __call__(self, id):
        return self[id].values.tolist()

    def __getitem__(self, item):
        return self.feature_df.iloc[item]



# coding: utf-8
# create by tongshiwei on 2019/4/11

"""
自定义的配置文件及对应的解析工具包。目的是为了更方便、快速地进行文件参数配置与解析。
"""
import argparse
import inspect
import json
import logging
import os
import re
import sys
from copy import deepcopy
from typing import List, Dict, Iterable

import toml
import yaml
from longling import wf_open
from longling.lib.path import path_append

__all__ = [
    "CLASS_EXCLUDE_NAMES", "get_class_var",
    "get_parsable_var", "load_configuration",
    "var2exp", "path_append",
    "Configuration", "ConfigurationParser",
    "Formatter", "ParserGroup", "is_classmethod"
]

CLASS_EXCLUDE_NAMES = set(vars(object).keys()) | {
    '__module__', '__main__', '__dict__', '__weakref__',
}


def get_class_var(class_obj, exclude_names: (set, None) = None, get_vars=None) -> dict:
    """
    Update in v1.3.18

    获取某个类的所有属性的变量名及其值

    Examples
    --------
    >>> class A(object):
    ...     att1 = 1
    ...     att2 = 2
    >>> get_class_var(A)
    {'att1': 1, 'att2': 2}
    >>> get_class_var(A, exclude_names={"att1"})
    {'att2': 2}
    >>> class B(object):
    ...     att3 = 3
    ...     att4 = 4
    ...     @staticmethod
    ...     def excluded_names():
    ...         return {"att4"}
    >>> get_class_var(B)
    {'att3': 3}

    Parameters
    ----------
    class_obj:
        类或类实例。需要注意两者的区别。
    exclude_names:
        需要排除在外的变量名。也可以通过在类定义 excluded_names 方法来指定要排除的变量名。
    get_vars

    Returns
    -------
    class_var:
        类内属性变量名及值

    """
    default_excluded_names = CLASS_EXCLUDE_NAMES if exclude_names is None \
        else CLASS_EXCLUDE_NAMES | exclude_names
    excluded_names = default_excluded_names if not hasattr(
        class_obj, "excluded_names"
    ) else class_obj.excluded_names() | default_excluded_names

    variables = {}
    if get_vars is not False:  # True or None
        variables.update({
            k: v for k, v in vars(class_obj).items() if
            not inspect.isroutine(v) and k not in excluded_names
        })
    if get_vars is not True:  # False or None
        for k in dir(class_obj):
            if hasattr(class_obj, k):
                v = getattr(class_obj, k)
                if not inspect.isroutine(v) and k not in excluded_names:
                    variables[k] = v
    return variables


def parse_params(params: dict, parse_functions: (dict, None) = None) -> dict:
    """
    参数后处理，根据parse_functions中定义的转换函数，对params中的变量值进行转换。
    例如，可用于将无法直接dump的变量转换成可以dump的变量。

    Examples
    --------
    >>> params = {"a": 1}
    >>> parse_functions = {"a": lambda x: x + x}
    >>> parse_params(params, parse_functions)
    {'a': 2}
    >>> params = {"a": 1}
    >>> parse_params(params)
    {'a': 1}

    """
    if parse_functions is not None:
        for name, func in parse_functions.items():
            if name in params:
                params[name] = func(params[name])
    return params


def get_parsable_var(class_obj, parse_exclude: set = None, dump_parse_functions=None, get_vars=True):
    """获取所有可以被解析的参数及其值，可以使用dump_parse_functions来对不可dump的值进行转换"""
    params = get_class_var(class_obj, exclude_names=parse_exclude, get_vars=get_vars)
    return parse_params(params, dump_parse_functions)


def load_configuration(fp, file_format="json", load_parse_function=None):
    """
    装载配置文件

    Updated in version 1.3.16

    Parameters
    ----------
    fp
    file_format
    load_parse_function

    """
    if file_format == "json":
        params = json.load(fp)
    elif file_format == "toml":
        params = toml.load(fp)
    elif file_format == "yaml":
        params = yaml.load(fp, Loader=yaml.FullLoader)
    else:
        raise TypeError("Unsupported file format: %s, only `json`, `toml` and `yaml` are supported" % file_format)
    return parse_params(params, load_parse_function)


def var2exp(var_str, env_wrap=lambda x: x):
    """
    将含有 $ 标识的变量转换为表达式

    Parameters
    ----------
    var_str
    env_wrap

    Examples
    --------
    >>> root = "dir"
    >>> dataset = "d1"
    >>> eval(var2exp("$root/data/$dataset"))
    'dir/data/d1'
    """
    var_str = str(var_str)

    pattern = re.compile(r"\$(\w+)")

    env_vars = pattern.findall(var_str)

    exp = """str(r"%s").format(%s)""" % (
        re.sub(r"\$\w+", "{}", var_str), ",".join(
            [env_wrap(env_var) for env_var in env_vars]
        )
    )
    return exp


def is_classmethod(method):
    """

    Parameters
    ----------
    method

    Returns
    -------

    Examples
    --------
    >>> class A:
    ...     def a(self):
    ...         pass
    ...     @staticmethod
    ...     def b():
    ...         pass
    ...     @classmethod
    ...     def c(cls):
    ...         pass
    >>> obj = A()
    >>> is_classmethod(obj.a)
    False
    >>> is_classmethod(obj.b)
    False
    >>> is_classmethod(obj.c)
    True
    >>> def fun():
    ...     pass
    >>> is_classmethod(fun)
    False
    """
    bound_to = getattr(method, '__self__', None)
    if not isinstance(bound_to, type):
        # must be bound to a class
        return False
    name = method.__name__
    for cls in bound_to.__mro__:
        descriptor = vars(cls).get(name)
        if descriptor is not None:
            return isinstance(descriptor, classmethod)
    return False  # pragma: no cover


class Configuration(object):
    """
    自定义的配置文件基类


    Examples
    --------
    >>> c = Configuration(a=1, b="example", c=[0,2], d={"a1": 3})
    >>> c.instance_var
    {'a': 1, 'b': 'example', 'c': [0, 2], 'd': {'a1': 3}}
    >>> c.default_file_format()
    'json'
    >>> c.get("a")
    1
    >>> c.get("e") is None
    True
    >>> c.get("e", 0)
    0
    >>> c.update(e=2)
    >>> c["e"]
    2
    """

    def __init__(self, logger=logging, **kwargs):
        self.logger = logger

        params = self.class_var

        params.update(**kwargs)
        for param, value in params.items():
            # all class variables will be contained in instance variables
            setattr(self, "%s" % param, value)

    def deep_update(self, **kwargs):
        for param, value in kwargs.items():
            setattr(self, "%s" % param, value)

    def _update(self, **kwargs):
        self.deep_update(**kwargs)

    def update(self, **kwargs):
        params = self.instance_var
        params.update(**kwargs)
        self._update(**params)

    @classmethod
    def vars(cls):
        # todo: check why when deepcopy is missed, two instance will have same id for class_var
        return deepcopy(get_class_var(cls))

    @property
    def class_var(self):
        """
        获取所有设定的参数

        Returns
        -------
        parameters: dict
            all variables used as parameters
        """
        return self.vars()

    @classmethod
    def pvars(cls):
        return get_parsable_var(
            cls,
            parse_exclude=None,
            dump_parse_functions=None,
        )

    @property
    def parsable_var(self):
        """
        获取可以进行命令行设定的参数

        Returns
        -------
        store_vars: dict
            可以进行命令行设定的参数
        """
        return get_parsable_var(
            self,
            parse_exclude=None,
            dump_parse_functions=None,
        )

    @property
    def instance_var(self):
        return get_parsable_var(
            self,
            parse_exclude=None,
            dump_parse_functions=None,
        )

    @classmethod
    def load_parse_function(cls):
        return None

    @classmethod
    def default_file_format(cls):
        return "json"

    @classmethod
    def load_cfg(cls, cfg_path, file_format=None, **kwargs):
        """从配置文件中装载配置参数"""
        file_format = file_format if file_format is not None else cls.default_file_format()

        with open(cfg_path) as f:
            params = load_configuration(
                f, file_format=file_format, load_parse_function=cls.load_parse_function()
            )
        params.update(kwargs)
        return params

    @classmethod
    def load(cls, cfg_path, file_format=None, **kwargs):
        """
        从配置文件中装载配置类

        Updated in version 1.3.16
        """
        file_format = file_format if file_format is not None else cls.default_file_format()
        return cls(**cls.load_cfg(cfg_path, file_format=file_format, **kwargs))

    def dump(self, cfg_path: str, override=True, file_format=None):
        """
        将配置参数写入文件

        Updated in version 1.3.16

        Parameters
        ----------
        cfg_path: str
        override: bool
        file_format: str
        """
        if os.path.isfile(cfg_path) and not override:
            self.logger.warning(
                "file %s existed, dump aborted" % os.path.abspath(cfg_path)
            )
            return
        self.logger.info(
            "writing configuration parameters to %s" % os.path.abspath(cfg_path)
        )
        file_format = file_format if file_format is not None else self.default_file_format()

        with wf_open(cfg_path) as wf:
            if file_format == "json":
                json.dump(self.parsable_var, wf, indent=2)
            elif file_format == "toml":
                toml.dump(self.parsable_var, wf)
            elif file_format == "yaml":
                yaml.dump(self.parsable_var, wf)
            else:
                raise TypeError(
                    "Unsupported file format: %s, only `json`, `toml` and `yaml` are supported" % file_format
                )

    def __str__(self):
        string = []
        for k, v in vars(self).items():
            string.append("%s: %s" % (k, v))
        return "\n".join(string)

    def __repr__(self):
        return str(self)

    def __getitem__(self, item):
        return getattr(self, item)

    def __contains__(self, item):
        return item in vars(self)

    def get(self, k, default=None):
        if k in self:
            return self[k]
        else:
            return default

    def items(self):
        return {k: v for k, v in self.parsable_var.items()}

    @classmethod
    def run_time_variables(cls):
        return {'logger'}

    @classmethod
    def excluded_names(cls):
        """
        获取非参变量集

        Returns
        -------
        exclude names set: set
            所有非参变量
        """
        return CLASS_EXCLUDE_NAMES | {
            'class_var', 'parsable_var', 'instance_var', 'items', 'load', 'dump', 'help_info', 'get', 'update',
            'deep_update'
        } | cls.run_time_variables()

    @classmethod
    def help_info(cls) -> dict:
        return {}


# class InheritableConfiguration(Configuration):
#     """
#     Examples
#     -------
#     >>> class Parent(Configuration):
#     ...     a = 1
#     ...     b = 2
#     >>> class Child1(Parent):
#     ...     a = 0
#     ...     c = 3
#     # >>> c1 = Child1()
#     # >>> c1.parsable_var
#     >>> class Child2(Parent, InheritableConfiguration):
#     ...     c = 3
#     >>> c2 = Child2()
#     >>> c2.parsable_var
#
#     """
#     @classmethod
#     def vars(cls):
#         return get_class_var(cls, get_vars=False)
#
#     @property
#     def parsable_var(self):
#         """
#         获取可以进行命令行设定的参数
#
#         Returns
#         -------
#         store_vars: dict
#             可以进行命令行设定的参数
#         """
#         return get_parsable_var(
#             self,
#             parse_exclude=None,
#             dump_parse_functions=None,
#             get_vars=False
#         )


def value_parse(value):
    r"""
    将含有关键字的字符串转换为关键字指定的类型
    支持的关键字类型为python的基础数据类型:
    int, float, dict, list, set, tuple, None

    Parameters
    ----------
    value: str
        字符串值

    Examples
    --------
    >>> value_parse("int(1.0)")
    1
    >>> value_parse("float(1.0)")
    1.0
    >>> value_parse("dict(a=1, b=2.0, c='d')")
    {'a': 1, 'b': 2.0, 'c': 'd'}
    >>> value_parse("list([1, 2, 3, 4])")
    [1, 2, 3, 4]
    >>> value_parse("set([1, 1, 2, 3])")
    {1, 2, 3}
    >>> value_parse("tuple([1, 2, 3])")
    (1, 2, 3)
    >>> value_parse("None") is None
    True
    """
    if re.findall(r"(int|float|dict|list|set|tuple|bool)\(.*\)|None", value):
        value = eval(value)
    return value


def parse_dict_string(string):
    """
    解析字典字符串

    Examples
    --------
    >>> parse_dict_string("a=1;b=int(1);c=list([1]);d=None")
    {'a': '1', 'b': 1, 'c': [1], 'd': None}
    >>> parse_dict_string("")
    """
    if not string:
        return None
    else:
        name_value_items = [
            item.strip().split("=", 1) for item in string.strip().rstrip(";").split(";")
        ]
        return {
            name_value[0]: value_parse(name_value[1])
            for name_value in name_value_items
        }


def args_zips(args=None, defaults=None):
    """组合参数及默认值"""
    args = [] if args is None else args
    defaults = [] if defaults is None else defaults
    assert len(args) >= len(defaults)
    zipped_args = []
    zipped_args.extend([
        {"name": arg} for arg in args[:len(args) - len(defaults)]
    ])
    zipped_args.extend([
        {"name": arg, "default": default} for arg, default in zip(
            args[len(args) - len(defaults):], defaults
        )
    ])
    return zipped_args


class ConfigurationParser(argparse.ArgumentParser):
    """
    Update in v1.3.18

    配置文件解析类，可用于构建cli工具。该类首先读入所有目标配置文件类class_obj的所有类属性，解析后生成命令行。
    普通属性参数使用 "--att_name att_value" 来读入。另外提供一个额外参数标记 ‘--kwargs’ 来读入可选参数。
    可选参数格式为 ::

        --kwargs key1=value1;key2=value2;...

    首先生成一个解析类 ::

        cli_parser = ConfigurationParser(Configuration)

    除了解析已有配置文件外，解析类还可以进一步添加函数来生成子解析器 ::

        cli_parser = ConfigurationParser($function)

    或者 ::

        cli_parser = ConfigurationParser([$function1, $function2])

    用以下三种解析方式中任意一种来解析参数：

    * 命令行模式 ::

        cli_parser()

    * 字符串传参模式 ::

        cli_parser('$parameter1 $parameters ...')

    * 列表传参模式 ::

        cli_parser(["--a", "int(1)", "--b", "int(2)"])

    Notes
    -----
    包含以下关键字的字符串会在解析过程中进行类型转换

    int, float, dict, list, set, tuple, None

    Parameters
    ----------
    class_type:
        类。注意是类，不是类实例。
    excluded_names:
        类中不进行解析的变量名集合
    commands:
        待解析的命令函数

    Examples
    --------
    >>> class TestC(Configuration):
    ...     a = 1
    ...     b = 2
    >>> def test_f1(k=1):
    ...     return k
    >>> def test_f2(h=1):
    ...      return h
    >>> def test_f3(m):
    ...      return m
    >>> parser = ConfigurationParser(TestC)
    >>> parser("--a 1 --b 2")
    {'a': '1', 'b': '2'}
    >>> ConfigurationParser.get_cli_cfg(TestC)
    {'a': 1, 'b': 2}
    >>> parser(["--a", "1", "--b", "int(1)"])
    {'a': '1', 'b': 1}
    >>> parser(["--a", "1", "--b", "int(1)", "--kwargs", "c=int(3);d=None"])
    {'a': '1', 'b': 1, 'c': 3, 'd': None}
    >>> parser.add_command(test_f1, test_f2, test_f3)
    >>> parser(["test_f1"])
    {'a': 1, 'b': 2, 'k': 1, 'subcommand': 'test_f1'}
    >>> parser(["test_f2"])
    {'a': 1, 'b': 2, 'h': 1, 'subcommand': 'test_f2'}
    >>> parser(["test_f3", "3"])
    {'a': 1, 'b': 2, 'm': '3', 'subcommand': 'test_f3'}
    >>> parser = ConfigurationParser(TestC, commands=[test_f1, test_f2])
    >>> parser(["test_f1"])
    {'a': 1, 'b': 2, 'k': 1, 'subcommand': 'test_f1'}
    >>> class TestCC:
    ...     c = {"_c": 1, "_d": 0.1}
    >>> parser = ConfigurationParser(TestCC)
    >>> parser("--c _c=int(3);_d=float(0.3)")
    {'c': {'_c': 3, '_d': 0.3}}
    >>> class TestCls:
    ...     def a(self, a=1):
    ...         return a
    ...     @staticmethod
    ...     def b(b=2):
    ...         return b
    ...     @classmethod
    ...     def c(cls, c=3):
    ...         return c
    >>> parser = ConfigurationParser(TestCls, commands=[TestCls.b, TestCls.c])
    >>> parser("b")
    {'b': 2, 'subcommand': 'b'}
    >>> parser("c")
    {'c': 3, 'subcommand': 'c'}
    """

    def __init__(self, class_type, excluded_names: (set, None) = None,
                 commands=None,
                 *args,
                 params_help=None, commands_help=None, override_help=False,
                 **kwargs):
        excluded_names = {
            'logger'
        } if excluded_names is None else excluded_names

        self.__proto_type = argparse.ArgumentParser(add_help=False)

        params = {k: v for k, v in get_class_var(class_type).items()}
        _params_help = {}
        _params_help.update(class_type.help_info() if hasattr(class_type, 'help_info') else {})
        _params_help.update({} if params_help is None else params_help)
        params_help = _params_help

        def get_help_info(default_help_info, param_name=None):
            if not params_help or param_name not in params_help:
                return default_help_info
            else:
                if override_help:
                    return "%s" % params_help[param_name]
                else:
                    return "%s, %s" % (params_help[param_name], default_help_info)

        for param, value in params.items():
            if param in excluded_names:  # pragma: no cover
                continue
            if isinstance(value, dict):
                format_tips = ", dict variables, " \
                              "use format: <key>=<value>(;<key>=<value>)"
                self.__proto_type.add_argument(
                    '--%s' % param,
                    help=get_help_info('set %s, default is %s%s' % (
                        param, value, format_tips
                    ), param),
                    default=value,
                    type=parse_dict_string,
                )
            else:
                self.__proto_type.add_argument(
                    '--%s' % param,
                    help=get_help_info('set %s, default is %s' % (param, value), param),
                    default=value,
                    type=value_parse,
                )
        self.__proto_type.add_argument(
            '--kwargs', required=False,
            help=r"add extra argument here, "
                 r"use format: <key>=<value>(;<key>=<value>)"
        )
        self.sub_command_parsers = None
        super(ConfigurationParser, self).__init__(parents=[self.__proto_type], *args, **kwargs)

        if commands is not None:
            assert isinstance(commands, Iterable)
            self.add_command(*commands, help_info=commands_help)

    def add_command(self, *commands, help_info: (List[Dict], dict, str, None) = None):
        """批量添加子命令解析器"""
        if isinstance(help_info, list):
            for command, _help_info in zip(commands, help_info):
                self._add_subcommand(self.func_spec(command), _help_info)
        else:
            for command in commands:
                self._add_subcommand(self.func_spec(command), help_info)

    @staticmethod
    def parse(arguments):
        """参数后解析"""
        arguments = vars(arguments) if not isinstance(arguments, dict) else arguments
        args_dict = dict()
        for k, v in arguments.items():
            if k == "kwargs" and v:
                kwargs = parse_dict_string(v)
                if kwargs:
                    args_dict.update(kwargs)
            elif k == "kwargs":
                pass
            else:
                args_dict[k] = v
        return args_dict

    @classmethod
    def get_cli_cfg(cls, params_class) -> dict:
        """获取默认配置参数"""
        params_parser = cls(params_class)
        kwargs = params_parser.parse_args([])
        kwargs = params_parser.parse(kwargs)
        return kwargs

    def __call__(self, args: (str, list, None) = None) -> dict:
        if isinstance(args, str):
            args = args.split(" ")
        kwargs = self.parse_args(args)
        kwargs = self.parse(kwargs)
        return kwargs

    def _add_subcommand(self, command_parameters, help_info=None):
        """添加子命令解析器"""

        def get_help_info(default_help_info, command=None):
            if isinstance(help_info, dict):
                return help_info.get(command, default_help_info)
            elif command is None:  # pragma: no cover
                return default_help_info
            else:
                return command

        if self.sub_command_parsers is None:
            self.sub_command_parsers = self.add_subparsers(
                parser_class=argparse.ArgumentParser,
                help='help for sub-command'
            )
        subparsers = self.sub_command_parsers
        subcommand, parameters = command_parameters
        subparser = subparsers.add_parser(
            subcommand, help=get_help_info("%s help" % subcommand, subcommand), parents=[self.__proto_type],
        )
        subparser.set_defaults(subcommand=subcommand)
        for parameter in parameters:
            if "default" in parameter:
                subparser.add_argument(
                    "--%s" % parameter["name"],
                    default=parameter["default"],
                    required=False,
                    type=value_parse,
                    help="set %s, default is %s" % (
                        parameter["name"], parameter["default"]
                    )
                )
            else:
                subparser.add_argument(
                    parameter["name"],
                    type=value_parse,
                    help="set %s" % parameter["name"]
                )

    @staticmethod
    def func_spec(f):
        """获取函数参数表"""
        argspec = inspect.getfullargspec(f)
        argspec_args = argspec.args[1:] if is_classmethod(f) else argspec.args

        return f.__name__, args_zips(
            argspec_args, argspec.defaults
        ) + args_zips(
            argspec.kwonlyargs,
            argspec.kwonlydefaults.values() if argspec.kwonlydefaults else []
        )


class ParserGroup(argparse.ArgumentParser):
    """
    >>> class TestC(Configuration):
    ...     a = 1
    ...     b = 2
    >>> def test_f1(k=1):
    ...     return k
    >>> def test_f2(h=1):
    ...      return h
    >>> class TestC2(Configuration):
    ...     c = 3
    >>> parser1 = ConfigurationParser(TestC, commands=[test_f1])
    >>> parser2 = ConfigurationParser(TestC, commands=[test_f2])
    >>> pg = ParserGroup({"model1": parser1, "model2": parser2})
    >>> pg(["model1", "test_f1"])
    {'a': 1, 'b': 2, 'k': 1, 'subcommand': 'test_f1'}
    >>> pg("model2 test_f2")
    {'a': 1, 'b': 2, 'h': 1, 'subcommand': 'test_f2'}
    """

    def __init__(self, parsers: dict, prog=None,
                 usage=None,
                 description=None,
                 epilog=None, add_help=True):
        super(ParserGroup, self).__init__(
            prog=prog,
            usage=usage,
            description=description,
            epilog=epilog,
            add_help=add_help
        )
        self.parsers = parsers

    def parse_args(self, args=None):
        args = sys.argv[1:] if args is None else args
        if args[0] in {"-h", "--help"}:  # pragma: no cover
            return super(ParserGroup, self).parse_args(args)
        else:
            target_parser = self.parsers[args[0]]
            real_args = args[1:]
            return target_parser.parse_args(real_args)

    def __call__(self, args: (str, list, None) = None) -> (dict, None):
        if isinstance(args, str):
            args = args.split(" ")
        args = sys.argv[1:] if args is None else args
        if args[0] in {"-h", "--help"}:  # pragma: no cover
            self.parse_args(args)
            return
        kwargs = self.parse_args(args)
        target_parser = args[0]
        return self.parsers[target_parser].parse(kwargs)

    def format_help(self) -> str:
        # hacking the original method
        formatter = self._get_formatter()

        # usage
        formatter.add_usage(self.usage, self._actions,
                            self._mutually_exclusive_groups)

        # description
        formatter.add_text(self.description)

        # positionals, optionals and user-defined groups
        for title, parser in self.parsers.items():
            formatter.start_section(title)
            formatter.add_text(parser.description)
            for action_group in parser._action_groups:
                formatter.start_section(action_group.title)
                formatter.add_text(action_group.description)
                formatter.add_arguments(action_group._group_actions)
                formatter.end_section()
            formatter.end_section()

        # epilog
        formatter.add_text(self.epilog)

        # determine help from format above
        return formatter.format_help()


class Formatter(object):
    """
    以特定格式格式化字符串

    Examples
    --------
    >>> formatter = Formatter()
    >>> formatter("hello world")
    'hello world'
    >>> formatter = Formatter("hello {}")
    >>> formatter("world")
    'hello world'
    >>> formatter = Formatter("hello {} v{:.2f}")
    >>> formatter("world", 0.2)
    'hello world v0.20'
    >>> formatter = Formatter("hello {1} v{0:.2f}")
    >>> formatter(0.2, "world")
    'hello world v0.20'
    >>> Formatter.format(0.2, "world", formatter="hello {1} v{0:.3f}")
    'hello world v0.200'
    """

    def __init__(self, formatter: (str, None) = None):
        self.formatter = formatter

    def __call__(self, *format_string):
        if self.formatter is None:
            assert len(format_string) == 1, \
                "formatter is None, the input should be a single value, now is %s, " \
                "which has %s value" % (format_string, len(format_string))
            return format_string[0]
        else:
            return self.formatter.format(*format_string)

    @staticmethod
    def format(*format_string, formatter=None):
        formatter = Formatter(formatter)
        return formatter(*format_string)



# coding: utf-8
# 2021/3/15 @ tongshiwei

def get_max_update(update_epoch, batches_per_epoch, warmup_steps=0):
    """

    Parameters
    ----------
    update_epoch: int
    batches_per_epoch: int
    warmup_steps: int

    Returns
    -------
    max_update: int
        maximum number of updates before the decay reaches final learning rate, which includs warmup steps
    Examples
    --------
    >>> get_max_update(10, 30)
    300
    >>> get_max_update(10, 30, 300)
    600
    """
    return update_epoch * batches_per_epoch + warmup_steps


def get_lr_factor(base_lr, step, max_update, final_lr=None, discount=None):
    """

    Parameters
    ----------
    base_lr
    step: int
        Changes the learning rate for every n updates.
    max_update: int
        maximum number of updates before the decay reaches final learning rate, which includs warmup steps
    final_lr:
        final learning rate after all steps

    discount

    Returns
    -------
    step: int
        Changes the learning rate for every n updates.
    factor:
    final_lr:
        final learning rate after all steps

    Examples
    --------
    >>> get_lr_factor(1, 10, 100, final_lr=1e-10)
    (10, 0.09999999999999999, 1e-10)
    >>> get_lr_factor(0.1, 10, 100, discount=1e-10)
    (10, 0.09999999999999999, 1.0000000000000001e-11)
    """
    final_lr, discount = get_final_lr(base_lr, final_lr, discount)
    return step, pow(discount, step / max_update), final_lr


def get_step(batches_per_epoch, epoch_update_freq=1):
    """
    Parameters
    ---------
    batches_per_epoch

    epoch_update_freq: int
        the frequence to change the learning rate (freq = 1 / (n times at one epoch) ).
        if epoch_update_freq <= 0, update at each step (batch)

    Returns
    -------
    step: int
        Changes the learning rate for every n updates.

    Examples
    -------
    >>> get_step(100, 1)
    100
    >>> get_step(100, 2)
    50
    """
    return batches_per_epoch // epoch_update_freq if epoch_update_freq >= 1 else 1


def get_final_lr(base_lr, final_lr=None, discount=None):
    """

    Parameters
    ----------
    base_lr
    final_lr:
        final learning rate after all steps
    discount

    Returns
    -------
    final_lr: int or float
        final learning rate after all steps
    discount:

    Examples
    --------
    >>> final_lr, discount = get_final_lr(0.1, 0.01)
    >>> round(final_lr, 2)
    0.01
    >>> round(discount, 2)
    0.1
    >>> final_lr, discount = get_final_lr(0.1, discount=0.1)
    >>> round(final_lr, 2)
    0.01
    >>> round(discount, 2)
    0.1
    """
    assert final_lr or discount
    if final_lr:
        return final_lr, final_lr / base_lr
    elif discount:
        return base_lr * discount, discount



# coding: utf-8
# 2020/8/16 @ tongshiwei
import logging
import os

import mxnet as mx
from mxnet.initializer import Initializer, Xavier, Uniform, Normal


def net_initialize(
        net, model_ctx,
        initializer: (str, Initializer, dict, list) = mx.init.Xavier(),
        select=None, logger=logging, verbose=False, force_reinit=False
):
    """
    初始化网络参数

    Parameters
    ----------
    net
    model_ctx: mx.cpu or mx.gpu
    initializer: str, Initializer, dict or list, tuple
    select
    logger
    verbose : bool, default False
        Whether to verbosely print out details on initialization.
    force_reinit : bool, default False
        Whether to force re-initialization if parameter is already initialized.
    Notes
    ------
    The developer who modify this document should simultaneously modify the related function in glue

    Examples
    --------
    >>> import mxnet as mx
    >>> from mxnet import gluon
    >>> emb = gluon.nn.Embedding(2, 3)
    >>> net_initialize(emb, mx.cpu())
    >>> emb.weight.data()
    <BLANKLINE>
    [[0.10694504 0.2034123  0.4714563 ]
     [0.7542485  0.2251432  0.7842196 ]]
    <NDArray 2x3 @cpu(0)>
    >>> emb1 = gluon.nn.Embedding(2, 3)
    >>> net_initialize(emb1, mx.cpu(), initializer=mx.init.Xavier())
    >>> emb1.weight.data()
    <BLANKLINE>
    [[ 0.09833419  0.76079047 -0.16726398]
     [ 0.27071452  0.319638   -0.25330698]]
    <NDArray 2x3 @cpu(0)>
    >>> class EmbNet(gluon.nn.HybridBlock):
    ...     def __init__(self, prefix=None, params=None):
    ...         super(EmbNet, self).__init__(prefix, params)
    ...         with self.name_scope():
    ...             self.emb = gluon.nn.Embedding(2, 3)
    ...             self.linear = gluon.nn.Dense(4)
    ...     def hybrid_forward(self, F, x, *args, **kwargs):
    ...         return self.linear(self.emb(x))
    >>> net = EmbNet()
    >>> from longling.ML.DL import BLOCK_EMBEDDING
    >>> net_initialize(net, mx.cpu(), initializer={BLOCK_EMBEDDING: "xaiver", ".*embedding": "uniform"})
    >>> net(mx.nd.array([0, 1]))
    <BLANKLINE>
    [[ 0.03268543 -0.00860071  0.04774952  0.00056277]
     [-0.00648303 -0.03121923 -0.04578817 -0.08059631]]
    <NDArray 2x4 @cpu(0)>
    >>> net1 = EmbNet()
    >>> net_initialize(net1, mx.cpu(), initializer=["xaiver", "uniform"], select=[BLOCK_EMBEDDING, ".*embedding"])
    >>> net1(mx.nd.array([0, 1]))  # doctest: +ELLIPSIS
    <BLANKLINE>
    [[-0.0896... -0.0179... -0.0156... -0.0136...]
     [ 0.0033...  0.0255...  0.0111...  0.0446...]]
    <NDArray 2x4 @cpu(0)>
    >>> net_initialize(net1, mx.cpu(), initializer=[(BLOCK_EMBEDDING, "xaiver"), (".*embedding", "uniform")],
    ...     force_reinit=True)
    >>> net1(mx.nd.array([0, 1]))  # doctest: +ELLIPSIS
    <BLANKLINE>
    [[ 0.0153...  0.0266... -0.0466...  0.0291...]
     [-0.0362...  0.0063...  0.0227... -0.0212...]]
    <NDArray 2x4 @cpu(0)>
    """
    if isinstance(initializer, str):
        initializer = {
            "xaiver": Xavier(),
            "uniform": Uniform(),
            "normal": Normal()
        }[initializer]
    elif isinstance(initializer, dict):
        for _select, _initializer in initializer.items():
            net_initialize(net, model_ctx=model_ctx, initializer=_initializer, select=_select, logger=logger,
                           verbose=verbose, force_reinit=force_reinit)
        return
    elif isinstance(initializer, (list, tuple)):
        if select is not None:
            assert len(select) == len(initializer)
            for _select, _initializer in zip(select, initializer):
                net_initialize(net, model_ctx=model_ctx, initializer=_initializer, select=_select, logger=logger,
                               verbose=verbose, force_reinit=force_reinit)
        else:
            for _select, _initializer in initializer:
                net_initialize(net, model_ctx=model_ctx, initializer=_initializer, select=_select, logger=logger,
                               verbose=verbose, force_reinit=force_reinit)
        return
    elif initializer is None or isinstance(initializer, Initializer):
        pass
    else:
        raise TypeError("initializer should be either str or Initializer, now is", type(initializer))

    logger.info("initializer: %s, select: %s, ctx: %s" % (initializer, select, model_ctx))
    net.collect_params(select).initialize(initializer, ctx=model_ctx, verbose=verbose, force_reinit=force_reinit)


def load_net(filename, net, ctx=mx.cpu(), allow_missing=False,
             ignore_extra=False):
    """
    Load the existing net parameters

    Parameters
    ----------
    filename: str
        The model file
    net: HybridBlock
        The network which has been initialized or
        loaded from the existed model
    ctx: Context or list of Context
            Defaults to ``mx.cpu()``.
    allow_missing: bool
    ignore_extra: bool

    Returns
    -------
    The initialized net
    """
    # 根据文件名装载已有的网络参数
    if not os.path.isfile(filename):
        raise FileExistsError("%s does not exist" % filename)
    net.load_parameters(filename, ctx, allow_missing=allow_missing,
                        ignore_extra=ignore_extra)
    return net



# coding: utf-8
# create by tongshiwei on 2020-11-12

import re

import mxnet.ndarray as nd

__all__ = ["save_params"]


def save_params(filename, net, select=None):
    """
    New in version 1.3.16

    Notes
    ------
    Q: Why not use the `save_parameters` in `mxnet.gluon`.
    A: Because we want to use the `select` argument to only preserve those parameters we want
    (i.e., excluding those unnecessary parameters such as pretrained embedding).
    """
    params = net._collect_params_with_prefix()
    if select is not None:
        pattern = re.compile(select)
        params = {name: value for name, value in params.items() if
                  pattern.match(name)}
    arg_dict = {key: val._reduce() for key, val in params.items()}
    nd.save(filename, arg_dict)
    return filename



# coding: utf-8
# 2019/9/21 @ tongshiwei


def legal_input(__promt: str,
                __legal_input: set = None, __illegal_input: set = None, is_legal=None,
                __default_value: str = None):
    """
    To make sure the input legal, if the input is illegal, the user will be asked to retype the input.

    Input being illegal means the input either not in __legal_input or in __illegal_input.

    For the case that user do not type in anything, if the __default_value is set, the __default_value will be used.
    Else, the input will be treated as illegal.

    Parameters
    ----------
    __promt: str
        tips to be displayed
    __legal_input: set
        the input should be in the __legal_input set
    __illegal_input: set
        the input should be not in the __illegal_input set
    is_legal: function
        the function used to judge whether the input is legal, by default, we use the inner __is_legal function
    __default_value: str
        default value when user type in nothing

    """
    inp = input(__promt)
    __legal_input = __legal_input if __legal_input else set()
    __illegal_input = __illegal_input if __illegal_input else set()

    def __is_legal(__inp):
        if (not __legal_input or __inp in __legal_input) and __inp not in __illegal_input:
            # whether the input either not in __legal_input or in __illegal_input
            return True
        elif not inp and __default_value:
            # If the __default_value is set, the __default_value will be used.
            # Else, the input will be treated as illegal.
            return True
        return False

    is_legal = is_legal if is_legal else __is_legal

    while not is_legal(inp):
        # the user will be asked to retype the input.
        inp = input(__promt)
    if not inp and __default_value:
        return __default_value
    return inp


def default_legal_input(tip, __legal_input: set = None, __illegal_input: set = None, is_legal=None,
                        __default_value: str = None, cursor="<"):
    __promt = tip

    legal_tip = []

    if __legal_input is not None:
        legal_tip.append("/".join(__legal_input))

    if __default_value is not None:
        legal_tip.append("default is %s" % __default_value)

    if legal_tip:
        __promt = tip + " (%s)" % ", ".join(legal_tip)

    __promt += " %s " % cursor
    return legal_input(__promt, __legal_input=__legal_input, __illegal_input=__illegal_input, is_legal=is_legal,
                       __default_value=__default_value)


def binary_legal_input(tip, _y="y", _n="n", _default="y"):
    assert _default in [_y, _n, None]
    return True if legal_input("%s (%s/%s, default is %s) < " % (tip, _y, _n, _default),
                               __default_value=_default) == _y else False



__all__ = ['as_list', 'dict2pv', 'list2dict', 'get_dict_by_path', 'format_byte_sizeof', 'group_by_n', 'as_ordered_dict']

from collections import OrderedDict


def group_by_n(obj: list, n: int) -> list:
    """
    Examples
    --------
    >>> list_obj = [1, 2, 3, 4, 5, 6]
    >>> group_by_n(list_obj, 3)
    [[1, 2, 3], [4, 5, 6]]
    """
    return [obj[i: i + n] for i in range(0, len(obj), n)]


def as_list(obj) -> list:
    r"""A utility function that converts the argument to a list
    if it is not already.

    Parameters
    ----------
    obj : object
        argument to be converted to a list

    Returns
    -------
    list_obj: list
        If `obj` is a list or tuple, return it. Otherwise,
        return `[obj]` as a single-element list.

    Examples
    --------
    >>> as_list(1)
    [1]
    >>> as_list([1])
    [1]
    >>> as_list((1, 2))
    [1, 2]
    """
    if isinstance(obj, list):
        return obj
    elif isinstance(obj, tuple):
        return list(obj)
    else:
        return [obj]


def as_ordered_dict(dict_data: (dict, OrderedDict), index: (list, None) = None):
    """
    Examples
    --------
    >>> as_ordered_dict({0: 0, 2: 123, 1: 1})
    OrderedDict([(0, 0), (2, 123), (1, 1)])
    >>> as_ordered_dict({0: 0, 2: 123, 1: 1}, [2, 0, 1])
    OrderedDict([(2, 123), (0, 0), (1, 1)])
    >>> as_ordered_dict(OrderedDict([(2, 123), (0, 0), (1, 1)]))
    OrderedDict([(2, 123), (0, 0), (1, 1)])
    """

    if index is None and isinstance(dict_data, OrderedDict):
        return dict_data
    elif index is None and isinstance(dict_data, dict):
        return OrderedDict(dict_data)

    ret = OrderedDict()

    for key in index:
        ret[key] = dict_data[key]

    return ret


def list2dict(list_obj, value=None, dict_obj=None):
    """
    >>> list_obj = ["a", 2, "c"]
    >>> list2dict(list_obj, 10)
    {'a': {2: {'c': 10}}}
    """
    dict_obj = {} if dict_obj is None else dict_obj
    _dict_obj = dict_obj
    for e in list_obj[:-1]:
        if e not in _dict_obj:
            _dict_obj[e] = {}
        _dict_obj = _dict_obj[e]
    _dict_obj[list_obj[-1]] = value

    return dict_obj


def get_dict_by_path(dict_obj, path_to_node):
    """
    >>> dict_obj = {"a": {"b": {"c": 1}}}
    >>> get_dict_by_path(dict_obj, ["a", "b", "c"])
    1
    """
    _dict_obj = dict_obj
    for p in path_to_node:
        _dict_obj = _dict_obj[p]
    return _dict_obj


def dict2pv(dict_obj: dict, path_to_node: list = None):
    """
    >>> dict_obj = {"a": {"b": [1, 2], "c": "d"}, "e": 1}
    >>> path, value = dict2pv(dict_obj)
    >>> path
    [['a', 'b'], ['a', 'c'], ['e']]
    >>> value
    [[1, 2], 'd', 1]
    """
    paths = []
    values = []
    path_to_node = [] if path_to_node is None else path_to_node
    for key, value in dict_obj.items():
        if isinstance(value, dict):
            _paths, _values = dict2pv(dict_obj[key], path_to_node + [key])
            paths.extend(_paths)
            values.extend(_values)
        else:
            paths.append(path_to_node + [key])
            values.append(value)
    return paths, values


def get_all_subclass(cls):  # pragma: no cover
    subclass = set()

    def _get_all_subclass(cls, res_set):
        res_set.add(cls)
        for sub_cls in cls.__subclasses__():
            _get_all_subclass(sub_cls, res_set)

    _get_all_subclass(cls, res_set=subclass)
    return subclass


def format_sizeof(num, suffix='', divisor=1000):
    """
    Code from tqdm

    Formats a number (greater than unity) with SI Order of Magnitude
    prefixes.

    Parameters
    ----------
    num  : float
        Number ( >= 1) to format.
    suffix  : str, optional
        Post-postfix [default: ''].
    divisor  : float, optional
        Divisor between prefixes [default: 1000].

    Returns
    -------
    out  : str
        Number with Order of Magnitude SI unit postfix.

    Examples
    --------
    >>> format_sizeof(800000)
    '800K'
    >>> format_sizeof(40000000000000)
    '40.0T'
    >>> format_sizeof(1000000000000000000000000000)
    '1000.0Y'
    """
    for unit in ['', 'K', 'M', 'G', 'T', 'P', 'E', 'Z']:
        if abs(num) < 999.5:
            if abs(num) < 99.95:
                if abs(num) < 9.995:
                    return '{0:1.2f}'.format(num) + unit + suffix
                return '{0:2.1f}'.format(num) + unit + suffix
            return '{0:3.0f}'.format(num) + unit + suffix
        num /= divisor
    return '{0:3.1f}Y'.format(num) + suffix


def format_byte_sizeof(num, suffix='B'):
    """
    Examples
    --------
    >>> format_byte_sizeof(1024)
    '1.00KB'
    """
    return format_sizeof(num, suffix, divisor=1024)


class RegisterNameExistedError(Exception):  # pragma: no cover
    pass


class Register(object):  # pragma: no cover
    def __init__(self):
        self.reg = set()

    def register(self, name):
        if name not in self.reg:
            self.reg.add(name)
            return True
        else:
            raise RegisterNameExistedError("name %s existed" % name)



# coding: utf-8
# 2021/2/12 @ tongshiwei

from longling.ML.DL import light_module as dlm
from .toolkit import loss_dict2tmt_torch_loss, get_trainer as _get_trainer, save_params as _save_params
from .tore import net_init as _net_init


def train(
        net, cfg, loss_function, trainer,
        train_data, test_data=None,
        params_save=False, dump_result=False, progress_monitor=None,
        *,
        fit_f, eval_f=None, get_net=None, get_loss=None, get_trainer=None, save_params=None,
        enable_hyper_search=False, reporthook=None, final_reporthook=None, primary_key=None,
        eval_epoch=1, initial_net=True, net_init=None, batch_lr_scheduler=None, epoch_lr_scheduler=None,
        loss_as_dict=False, verbose=None, dump_cfg=None, **cfg_kwargs
):
    if initial_net:
        net_init = _net_init if net_init is None else net_init
    else:
        net_init = None

    if trainer is None and get_trainer is None:
        get_trainer = _get_trainer

    if cfg.lr_params:
        assert batch_lr_scheduler or epoch_lr_scheduler

    save_params = _save_params if save_params is None else save_params

    dlm.train(
        net, cfg, loss_function, trainer, train_data, test_data, params_save, dump_result, progress_monitor,
        fit_f=fit_f, eval_f=eval_f,
        net_init=net_init, get_net=get_net, get_loss=get_loss, get_trainer=get_trainer,
        save_params=save_params,
        enable_hyper_search=enable_hyper_search, reporthook=reporthook,
        final_reporthook=final_reporthook,
        primary_key=primary_key, eval_epoch=eval_epoch, loss_dict2tmt_loss=loss_dict2tmt_torch_loss,
        batch_lr_scheduler=batch_lr_scheduler, epoch_lr_scheduler=epoch_lr_scheduler, loss_as_dict=loss_as_dict,
        verbose=verbose, dump_cfg=dump_cfg, **cfg_kwargs
    )



# coding: utf-8
# create by tongshiwei on 2019-8-30
import re

import torch
# from .lr_scheduler import get_lr_scheduler
from longling.ML.PytorchHelper.toolkit.lr_scheduler import get_lr_scheduler


def collect_params(net, select=None):
    """

    Parameters
    ----------
    net
    select

    Returns
    -------
    >>> class EmbedMLP(torch.nn.Module):
    ...     def __init__(self):
    ...         super(EmbedMLP, self).__init__()
    ...         self.embedding = torch.nn.Embedding(5, 3)
    ...         self.linear = torch.nn.Linear(3, 5)
    >>> net = EmbedMLP()
    >>> len(list(collect_params(net, None)))
    3
    >>> from longling.ML.DL import ALL_PARAMS, BLOCK_EMBEDDING
    >>> len(list(collect_params(net, ALL_PARAMS)))
    3
    >>> len(list(collect_params(net, BLOCK_EMBEDDING)))
    2
    >>> len(list(collect_params(net.linear)))
    2
    >>> len(list(collect_params(net.embedding)))
    1
    """
    if select is None:
        return net.parameters()
    pattern = re.compile(select)
    ret = [value for name, value in net.named_parameters() if pattern.match(name)]
    return ret


def get_trainer(net, optimizer, optimizer_params=None, lr_params=None, select=None) -> ...:
    """

    Parameters
    ----------
    net
    optimizer
    optimizer_params
    lr_params
    select

    Returns
    -------

    Examples
    >>> class EmbedMLP(torch.nn.Module):
    ...     def __init__(self):
    ...         super(EmbedMLP, self).__init__()
    ...         self.embedding = torch.nn.Embedding(5, 3)
    ...         self.linear = torch.nn.Linear(3, 5)
    >>> net = EmbedMLP()
    >>> optimizer = get_trainer(net, "sgd", optimizer_params={"lr": 0.01})
    >>> optimizer
    SGD (
    Parameter Group 0
        dampening: 0
        lr: 0.01
        momentum: 0
        nesterov: False
        weight_decay: 0
    )
    >>> len(optimizer.param_groups)
    1
    >>> sum([len(group["params"]) for group in optimizer.param_groups])
    3
    >>> from longling.ML.DL import BLOCK_EMBEDDING
    >>> optimizer = get_trainer(net, "sgd", optimizer_params={"lr": 0.01}, select=BLOCK_EMBEDDING)
    >>> len(optimizer.param_groups)
    1
    >>> sum([len(group["params"]) for group in optimizer.param_groups])
    2
    >>> optimizer, lr_scheduler = get_trainer(
    ...     net, "sgd",
    ...     optimizer_params={"lr": 0.01},
    ...     lr_params={"scheduler": "OneCycleLR", "max_lr": 1, "total_steps": 100}
    ... )
    >>> optimizer
    SGD (
    Parameter Group 0
        base_momentum: 0.85
        dampening: 0
        initial_lr: 0.04
        lr: 0.040000000000000036
        max_lr: 1
        max_momentum: 0.95
        min_lr: 4e-06
        momentum: 0.95
        nesterov: False
        weight_decay: 0
    )
    >>> lr_scheduler  # doctest: +ELLIPSIS
    <torch.optim.lr_scheduler.OneCycleLR...
    >>> get_trainer(net, "adam", optimizer_params={"lr": 0.01})
    Adam (
    Parameter Group 0
        amsgrad: False
        betas: (0.9, 0.999)
        eps: 1e-08
        lr: 0.01
        weight_decay: 0
    )
    >>> get_trainer(net, "adagrad", optimizer_params={"lr": 0.01})
    Adagrad (
    Parameter Group 0
        eps: 1e-10
        initial_accumulator_value: 0
        lr: 0.01
        lr_decay: 0
        weight_decay: 0
    )
    >>> get_trainer(net, "rmsprop", optimizer_params={"lr": 0.01})
    RMSprop (
    Parameter Group 0
        alpha: 0.99
        centered: False
        eps: 1e-08
        lr: 0.01
        momentum: 0
        weight_decay: 0
    )
    >>> get_trainer(net, "Rprop", optimizer_params={"lr": 0.01})
    Rprop (
    Parameter Group 0
        etas: (0.5, 1.2)
        lr: 0.01
        step_sizes: (1e-06, 50)
    )
    >>> import torch
    >>> get_trainer(net, torch.optim.SGD, optimizer_params={"lr": 0.01})
    SGD (
    Parameter Group 0
        dampening: 0
        lr: 0.01
        momentum: 0
        nesterov: False
        weight_decay: 0
    )
    """
    assert isinstance(optimizer, str) or issubclass(optimizer, torch.optim.Optimizer)
    parameters = collect_params(net, select)
    assert optimizer_params is None or isinstance(optimizer_params, dict)
    optimizer_params = {} if optimizer_params is None else optimizer_params

    if isinstance(optimizer, str):
        if optimizer == "sgd":
            optimizer_class = torch.optim.SGD
        elif optimizer == "adam":
            optimizer_class = torch.optim.Adam
        elif optimizer == "adagrad":
            optimizer_class = torch.optim.Adagrad
        elif optimizer == "rmsprop":
            optimizer_class = torch.optim.RMSprop
        else:
            optimizer_class = eval("torch.optim.%s" % optimizer)
    else:
        optimizer_class = optimizer

    optimizer = optimizer_class(parameters, **optimizer_params)
    if lr_params:
        lr_scheduler = get_lr_scheduler(optimizer=optimizer, **lr_params)
        return optimizer, lr_scheduler
    else:
        return optimizer



# coding: utf-8
# create by tongshiwei on 2019/4/12
import os

__all__ = ["Module"]

import torch

from ..toolkit.trainer import get_trainer
from ..toolkit.parallel import set_device
from longling.ML import DL


class Module(DL.Module):
    @staticmethod
    def load_net(filename):
        """
        Load the existing net parameters

        Parameters
        ----------
        filename: str
            The model file
        Returns
        -------
        The initialized net
        """
        # 根据文件名装载已有的网络参数
        if not os.path.isfile(filename):
            raise FileExistsError
        return torch.load(filename, map_location=lambda s, loc: s)

    def load(self, epoch):
        """"
        Load the existing net parameters

        Parameters
        ----------
        epoch: str or int
            The epoch which specify the model

        Returns
        -------
        The initialized net
        """
        # 根据起始轮次装载已有的网络参数
        filename = self.epoch_params_filename(epoch)
        return self.load_net(
            filename
        )

    def epoch_params_filename(self, epoch):
        raise NotImplementedError

    @staticmethod
    def get_trainer(net, optimizer='sgd', optimizer_params=None, select=None):
        """把优化器安装到网络上"""
        trainer = get_trainer(net, optimizer, optimizer_params, select)
        return trainer

    @staticmethod
    def save_params(filename, net: torch.nn.Module):
        torch.save(net.state_dict(), filename)

    @staticmethod
    def net_initialize(net, model_ctx="cpu", **kwargs):
        """初始化网络参数"""
        return set_device(net, model_ctx, **kwargs)



# coding: utf-8

__all__ = ["toc"]

from pathlib import Path


def toc(root="./", parent=None, is_last=False, criteria=None):
    """
    打印目录树

    show the tree of contents in the specified root directory

    Parameters
    ----------
    root
    parent
    is_last
    criteria

    References
    ----------
    https://stackoverflow.com/a/49912639

    """
    paths = DisplayablePath.make_tree(root, parent, is_last, criteria)
    for path in paths:
        print(path.displayable())


class DisplayablePath(object):
    display_filename_prefix_middle = '├──'
    display_filename_prefix_last = '└──'
    display_parent_prefix_middle = '    '
    display_parent_prefix_last = '│   '

    def __init__(self, path, parent_path, is_last):
        self.path = Path(str(path))
        self.parent = parent_path
        self.is_last = is_last
        if self.parent:
            self.depth = self.parent.depth + 1
        else:
            self.depth = 0

    @classmethod
    def make_tree(cls, root, parent=None, is_last=False, criteria=None):
        root = Path(str(root))
        criteria = criteria or cls._default_criteria

        displayable_root = cls(root, parent, is_last)
        yield displayable_root

        children = sorted(list(path
                               for path in root.iterdir()
                               if criteria(path)),
                          key=lambda s: str(s).lower())
        count = 1
        for path in children:
            is_last = count == len(children)
            if path.is_dir():
                yield from cls.make_tree(path,
                                         parent=displayable_root,
                                         is_last=is_last,
                                         criteria=criteria)
            else:
                yield cls(path, displayable_root, is_last)
            count += 1

    @classmethod
    def _default_criteria(cls, path):
        return True

    @property
    def display_name(self):
        if self.path.is_dir():
            return self.path.name + '/'
        return self.path.name

    def displayable(self):
        if self.parent is None:
            return self.display_name

        _filename_prefix = (self.display_filename_prefix_last
                            if self.is_last
                            else self.display_filename_prefix_middle)

        parts = ['{!s} {!s}'.format(_filename_prefix,
                                    self.display_name)]

        parent = self.parent
        while parent and parent.parent is not None:
            parts.append(self.display_parent_prefix_middle
                         if parent.is_last
                         else self.display_parent_prefix_last)
            parent = parent.parent

        return ''.join(reversed(parts))



# coding: utf-8
# create by tongshiwei on 2019/7/2
# PYTEST_DONT_REWRITE

__all__ = ["path_append", "file_exist", "abs_current_dir", "type_from_name", "parent_dir", "tmpfile"]

import os
from pathlib import PurePath
import uuid
import tempfile
from contextlib import contextmanager


def type_from_name(filename):
    """
    Examples
    --------
    >>> type_from_name("1.txt")
    '.txt'
    """
    return PurePath(filename).suffix


def path_append(path, *addition, to_str=False):
    """
    路径合并函数

    Examples
    --------
    .. code-block:: python

    path_append("../", "../data", "../dataset1/", "train", to_str=True)
    '../../data/../dataset1/train'

    Parameters
    ----------
    path: str or PurePath
    addition: list(str or PurePath)
    to_str: bool
        Convert the new path to str
    Returns
    -------

    """
    path = PurePath(path)
    if addition:
        for a in addition:
            path = path / a
    if to_str:
        return str(path)
    return path


def file_exist(filepath):
    """判断文件是否存在"""
    return os.path.isfile(filepath)


def abs_current_dir(filepath):
    """
    获取文件所在目录的绝对路径

    Example
    -------
    .. code ::

        abs_current_dir(__file__)

    """
    return os.path.abspath(os.path.dirname(filepath))


def parent_dir(filepath, level):
    _path = filepath
    for _ in range(level):
        _path = path_append(_path, "..")

    return _path


@contextmanager
def tmpfile(suffix=None, prefix=None, dir=None):
    """
    Create a temporary file, which will automatically cleaned after used (outside "with" closure).

    Examples
    --------

    .. code-block ::

        with tmpfile("test_tmp") as tmp:
            print(tmp)
            with open(tmp, mode="w") as wf:
                print("hello world", file=wf)
    """
    prefix = prefix if prefix is not None else tempfile.gettempprefix() + str(uuid.uuid4())[:6]
    filename = prefix + suffix if suffix is not None else prefix
    with tempfile.TemporaryDirectory(dir=dir) as tmpdir:
        yield os.path.join(tmpdir, filename)



# coding: utf-8
# 2020/4/13 @ tongshiwei


from .classification import classification_report
from .regression import regression_report
from .ranking import ranking_report
from .utils import POrderedDict



# coding: utf-8
# create by tongshiwei on 2019-8-30

from .trainer import get_trainer
from .optimizer_cfg import get_optimizer_cfg
from .loss import loss_dict2tmt_torch_loss
from .init import net_initialize, load_net
from .fit import fit_wrapper
from .eval import eval_wrapper
from .parallel import set_device
from .persistence import save_params
from .lr_scheduler import get_lr_scheduler



# coding: utf-8
# 2019/9/21 @ tongshiwei

import os
from longling.lib.path import path_append
from longling.Architecture.install_file import sphinx_conf, readthedocs, gitignore, dockerfile


def docs_proj(tar_dir, docs_params, docker_params=None, __gitignore=True, **kwargs):
    assert docs_params

    variables = {}
    variables.update(docs_params)

    docs_root = path_append(tar_dir, docs_params["docs_root"])
    if not os.path.exists(docs_root):
        os.makedirs(docs_root)
    sphinx_conf(tar_dir=docs_root, **docs_params)

    if docs_params["readthedocs"]:
        readthedocs(tar_dir=tar_dir)

    if __gitignore:
        gitignore("docs", tar_dir)

    if docker_params:  # pragma: no cover
        dockerfile(docs_params["docker_type"], **docker_params, tar_dir=docs_root)

    elif "docker_params" in docs_params:
        dockerfile(docs_params["docker_params"]["docker_type"], **docs_params["docker_params"], tar_dir=docs_root)



# coding: utf-8
# create by tongshiwei on 2020-11-17

from longling import path_append

__all__ = ["epoch_params_filename", "get_epoch_params_filepath", "params_filename", "get_params_filepath"]


def params_filename(model_name: str):
    """
    Examples
    --------
    >>> params_filename("CNN")
    'CNN.params'
    """
    return "%s.params" % model_name


def get_params_filepath(model_name: str, model_dir: str = "./"):
    """
    Examples
    --------
    >>> get_params_filepath("CNN")
    'CNN.params'
    """
    return path_append(model_dir, params_filename(model_name), to_str=True)


def epoch_params_filename(model_name: str, epoch: int):
    """
    Examples
    --------
    >>> epoch_params_filename("CNN", 10)
    'CNN-0010.params'
    """
    return "%s-%04d.params" % (model_name, epoch)


def get_epoch_params_filepath(model_name: str, epoch: int, model_dir: str = "./"):
    """
    Examples
    --------
    >>> get_epoch_params_filepath("CNN", 10)
    'CNN-0010.params'
    """
    return path_append(model_dir, epoch_params_filename(model_name, epoch), to_str=True)



# coding: utf-8
# create by tongshiwei on 2019-9-1
__all__ = ["net_viz"]

import logging

import mxnet as mx

from longling import path_append
from longling.ML.MxnetHelper.toolkit.viz import plot_network, VizError


# todo: optional, visualize the network
def net_viz(_net, _cfg, view_tag=False, **kwargs):  # pragma: no cover
    """visualization check, only support pure static network"""
    batch_size = _cfg.batch_size
    model_dir = _cfg.model_dir
    logger = kwargs.get(
        'logger',
        _cfg.logger if hasattr(_cfg, 'logger') else logging
    )

    try:
        viz_dir = path_append(model_dir, "plot/network")
        logger.info("visualization: file in %s" % viz_dir)
        from copy import deepcopy

        viz_net = deepcopy(_net)
        viz_shape = {'data': (batch_size,) + (1,)}
        x = mx.sym.var("data")
        sym = viz_net(x)
        plot_network(
            nn_symbol=sym,
            save_path=viz_dir,
            shape=viz_shape,
            node_attrs={"fixedsize": "false"},
            view=view_tag
        )
    except VizError as e:
        logger.error("error happen in visualization, aborted")
        logger.error(e)



# coding: utf-8
# created by tongshiwei on 17-12-17
import logging
import io
import os
import re
from distutils.core import setup

from setuptools import find_packages

test_deps = [
    'pytest>=4',
    'pytest-cov>=2.6.0',
    'pytest-flake8',
]

docs_deps = [
    'sphinx',
    'sphinx_rtd_theme',
    'nbsphinx',
    'm2r2',
    'Image',
    'recommonmark',
    'ipython',
    'sphinx_toggleprompt'
]

dev_deps = test_deps + docs_deps + [
    'setuptools>=40',
    'wheel',
    'twine'
]

dm_base_deps = [
    "pandas",
    "numpy",
    "matplotlib",
]

ml_base_deps = [
    "pandas",
    "numpy>= 1.16.5",
    "scipy",
    "scikit-learn>=0.23.2",
]
viz_deps = [
    "matplotlib",
    "tensorboardx",
    "tensorboard"
]

try:
    import mxnet

    mxnet_requires = []
except ModuleNotFoundError:
    mxnet_requires = ["mxnet"]
except Exception as e:
    mxnet_requires = []
    logging.error(e)

try:
    import torch

    ml_pytorch_deps = []
except ModuleNotFoundError:
    import sys

    if 5 <= sys.version_info[1]:
        ml_pytorch_deps = ["torch"]
    else:
        ml_pytorch_deps = []
        logging.warning("Current python version %s is not supported by pytorch", str(sys.version_info[:2]))
except Exception as e:
    ml_pytorch_deps = []
    logging.error(e)

ml_mx_deps = ["gluonnlp"] + mxnet_requires

spider_deps = [
    "requests",
    "rarfile",
    "bs4",
    "lxml",
    "urllib3"
]

dl_deps = ml_mx_deps + ml_pytorch_deps

ml_full_deps = ml_base_deps + dl_deps + ["nni>=1.8"] + viz_deps

full_deps = ml_full_deps + spider_deps


def read(*names, **kwargs):
    with io.open(
            os.path.join(os.path.dirname(__file__), *names),
            encoding=kwargs.get("encoding", "utf8")
    ) as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


VERSION = find_version('longling', '__init__.py')

setup(
    name='longling',
    version=VERSION,
    author='Sherlock, Shiwei Tong',
    author_email='tongsw@mail.ustc.edu.cn',
    include_package_data=True,
    python_requires='>=3.6',
    packages=find_packages(
        include=[
            "longling",
            "*.toolbox", "*.toolbox.*",
            "*.lib", "*.lib.*",
            "*.spider", "*.spider.*",
            "*.Architecture", "*.Architecture.*",
            "*.ML*",
        ],
        exclude=[
            "*.mx_example", "*.gluon_example*", "*.gluon_exp*",
            "*.mxnet_old*",
        ]
    ),
    entry_points={
        "console_scripts": [
            "glue = longling.ML.MxnetHelper.glue.glue:cli",
            "longling = longling.main:cli"
        ],
    },
    url='https://gitlab.com/tswsxk/longling.git',
    license='LICENSE.txt',
    description='This project aims to provide some handy toolkit functions to help construct the architecture.',
    long_description=open('README.txt', encoding="utf-8").read(),
    install_requires=[
        "pip",
        "tqdm",
        "fire",
        "PyYAML",
        "toml",
    ],
    extras_require={
        'test': test_deps + full_deps,
        'doc': docs_deps,
        'dev': dev_deps,
        'dm': dm_base_deps,
        'ml': ml_base_deps,
        'ml-viz': ml_base_deps + viz_deps,
        'viz': viz_deps,
        'mx': ml_mx_deps,
        'torch': ml_pytorch_deps,
        'dl': dl_deps,
        'ml-full': ml_full_deps,
        "spider": spider_deps,
        "full": full_deps
    },
    classifiers=[
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



# coding: utf-8
# Copyright @tongshiwei

from __future__ import absolute_import

from .ModelName import ModelName
from .ModelName.Module import Module, net_init, Configuration, fit_f

__all__ = ["ModelName", "Module", "net_init", "Configuration", "fit_f"]



# coding: utf-8
# create by tongshiwei on 2019-8-30

__all__ = ["length2mask", "get_sequence_mask", "sequence_mask"]

import torch
from torch import Tensor

from .utils import tensor2list


def length2mask(length, max_len, valid_mask_val, invalid_mask_val):
    """
    >>> data_len = [1, 2, 3]
    >>> length2mask(data_len, 5, 1, 0)
    tensor([[1, 0, 0, 0, 0],
            [1, 1, 0, 0, 0],
            [1, 1, 1, 0, 0]])
    """
    mask = []

    if isinstance(valid_mask_val, Tensor):
        valid_mask_val = tensor2list(valid_mask_val)
    if isinstance(invalid_mask_val, Tensor):
        invalid_mask_val = tensor2list(invalid_mask_val)
    if isinstance(length, Tensor):
        length = tensor2list(length)

    for _len in length:
        mask.append([valid_mask_val] * _len + [invalid_mask_val] * (max_len - _len))

    return torch.tensor(mask)


def get_sequence_mask(shape, sequence_length, axis=1):
    """
    Parameters
    ----------
    shape:
    sequence_length
    axis: int
        the length axis,
        if there is a batch axis, should be placed before length, like (batch, length, ...)
        instead of (length, batch, ...)

    Examples
    --------
    >>> data_shape = (2, 3, 4)
    >>> mask = get_sequence_mask(data_shape, [1, 3])
    >>> mask.shape
    torch.Size([2, 3, 4])
    >>> mask
    tensor([[[1., 1., 1., 1.],
             [0., 0., 0., 0.],
             [0., 0., 0., 0.]],
    <BLANKLINE>
            [[1., 1., 1., 1.],
             [1., 1., 1., 1.],
             [1., 1., 1., 1.]]])
    """
    assert axis <= len(shape)
    mask_shape = shape[axis + 1:]

    valid_mask_val = torch.ones(mask_shape)
    invalid_mask_val = torch.zeros(mask_shape)

    max_len = shape[axis]

    return length2mask(sequence_length, max_len, valid_mask_val, invalid_mask_val)


def sequence_mask(tensor: Tensor, sequence_length, axis=1):
    """
    Parameters
    ----------
    tensor: Tensor
    sequence_length
    axis: int
        the length axis,
        if there is a batch axis, should be placed before length, like (batch, length, ...)
        instead of (length, batch, ...)

    Returns
    -------
    masked_tensor: Tensor

    Examples
    --------
    >>> seq = torch.tensor([[1, 1, 2], [3, 1, 0]])
    >>> sequence_mask(seq, [2, 1])
    tensor([[1., 1., 0.],
            [3., 0., 0.]])
    >>> sequence_mask(seq, torch.tensor([2, 1]))
    tensor([[1., 1., 0.],
            [3., 0., 0.]])
    """
    mask = get_sequence_mask(tensor.shape, sequence_length, axis).to(tensor.device)
    return tensor * mask



# coding: utf-8
# Copyright @tongshiwei
from __future__ import absolute_import

import functools
import logging
import os

import mxnet as mx
from longling.ML.MxnetHelper.glue import module
from longling.ML.toolkit import EpochEvalFMT as Formatter
from tqdm import tqdm

from .configuration import Configuration
from .sym import get_net, fit_f, eval_f, net_init

__all__ = ["Module"]


class Module(module.Module):
    """
    模块模板
    train 修改流程

    # 1
    修改 __init__ 和 params 方法
    Module(....) 初始化一些通用的参数，比如模型的存储路径等

    # 2
    定义网络部分，命名为
    sym_gen(), 可以有多个
    也可以使用直接装载的方式生成已有网络

    # 3
    使用可视化方法进行检查网络是否搭建得没有问题
    可视化方法命名为 visualization

    # 4
    定义数据装载部分，可以有多个，命名为
    get_{train/test/data}_iter

    # 5
    定义训练相关的参数及交互提示信息

    # 6
    定义 net_initialize 部分
    定义训练部分

    # 7
    关闭输入输出流

    # optional
    定义 eval 评估模型方法
    定义 load 加载模型方法
    """

    def __init__(self, configuration):
        """

        Parameters
        ----------
        configuration: Configuration

        """
        # 初始化一些通用的参数
        self.cfg = configuration
        self.prefix = os.path.join(
            self.cfg.model_dir, self.cfg.model_name
        ) if hasattr(self.cfg, "model_dir") and hasattr(self.cfg, "model_name") else "model"
        self.logger = configuration.logger

    @functools.wraps(fit_f)
    def fit_f(self, *args, **kwargs):
        return fit_f(*args, **kwargs)

    @functools.wraps(get_net)
    def sym_gen(self, *args, **kwargs):
        return get_net(*args, **kwargs)

    @functools.wraps(net_init)
    def net_initialize(self, *args, **kwargs):
        return net_init(*args, **kwargs)

    def dump_configuration(self, filename=None):
        filename = filename if filename is not None \
            else os.path.join(self.cfg.cfg_path)
        self.cfg.dump(filename, override=True)
        return filename

    def epoch_params_filepath(self, epoch):
        return self.prefix + "-%04d.parmas" % epoch

    # 部分定义训练相关的方法
    @staticmethod
    @functools.wraps(module.Module.get_trainer)
    def get_trainer(
            net, optimizer='sgd', optimizer_params=None, lr_params=None,
            select=Configuration.train_select, logger=logging, *args, **kwargs
    ):
        return module.Module.get_trainer(
            net=net,
            optimizer=optimizer,
            optimizer_params=optimizer_params,
            lr_params=lr_params,
            select=select,
            logger=logger,
            *args,
            **kwargs,
        )

    def save_params(self, filename, net):
        module.Module.save_params(filename, net, select=self.cfg.save_select)

    def fit(
            self,
            net, begin_epoch, end_epoch, batch_size,
            train_data,
            trainer,
            loss_function,
            eval_data=None,
            ctx=mx.cpu(),
            toolbox=None,
            **kwargs
    ):
        """
        API for train

        Parameters
        ----------
        net: HybridBlock
            The network which has been initialized or loaded from
            the existed model
        begin_epoch: int
            The begin epoch of this train procession
        end_epoch: int
            The end epoch of this train procession
        batch_size: int
                The size of each batch
        train_data: Iterable
            The data used for this train procession,
            NOTICE: should have been divided to batches
        trainer:
            The trainer used to update the parameters of the net
        loss_function: dict of function
            The functions to compute the loss for the procession
            of back propagation
        eval_data: Iterable
            The data used for the evaluation at the end of each epoch,
            NOTICE: should have been divided to batches
            Default to ``None``
        ctx: Context or list of Context
            Defaults to ``mx.cpu()``.
        toolbox: dict or None
            Default to ``None``
        kwargs

        Returns
        -------

        """
        # 此方法可以直接使用
        return self.epoch_loop(
            net=net, begin_epoch=begin_epoch, end_epoch=end_epoch,
            batch_size=batch_size,
            train_data=train_data,
            trainer=trainer,
            loss_function=loss_function,
            test_data=eval_data,
            ctx=ctx,
            toolbox=toolbox,
            **kwargs
        )

    def epoch_loop(
            self,
            net, begin_epoch, end_epoch, batch_size,
            train_data,
            trainer,
            loss_function,
            test_data=None,
            ctx=mx.cpu(),
            toolbox=None,
            save_model=True,
            eval_every_n_epoch=1,
            **kwargs
    ):
        """
        此函数包裹批次训练过程，形成轮次训练过程

        Parameters
        ----------
        net: HybridBlock
            The network which has been initialized or loaded from
            the existed model
        begin_epoch: int
            The begin epoch of this train procession
        end_epoch: int
            The end epoch of this train procession
        batch_size: int
            The size of each batch
        train_data: Iterable
            The data used for this train procession,
            NOTICE: should have been divided to batches
        trainer:
            The trainer used to update the parameters of the net
        loss_function: dict of function
            The functions to compute the loss for the procession
            of back propagation
        test_data: Iterable
            The data used for the evaluation at the end of each epoch,
            NOTICE: should have been divided to batches
            Default to ``None``
        ctx: Context or list of Context
            Defaults to ``mx.cpu()``.
        toolbox: Toolbox
            Default to ``None``
        save_model: bool
            Whether save model
        eval_every_n_epoch: int
        kwargs
        """
        # 参数修改时需要同步修改 fit 函数中的参数
        # 定义轮次训练过程
        if toolbox is not None:
            formatter = toolbox.get('formatter')
        else:
            formatter = None

        for epoch in range(begin_epoch, end_epoch):
            batch_num, loss_values = self.batch_loop(
                net=net, epoch=epoch, batch_size=batch_size,
                train_data=train_data,
                trainer=trainer,
                loss_function=loss_function,
                ctx=ctx,
                toolbox=toolbox,
            )
            if hasattr(self.cfg, "lr_params") and self.cfg.lr_params \
                    and "update_params" in self.cfg.lr_params and self.cfg.end_epoch - self.cfg.begin_epoch - 1 > 0:
                self.cfg.logger.info("reset trainer")
                lr_params = self.cfg.lr_params.pop("update_params")
                lr_update_params = dict(
                    batches_per_epoch=batch_num,
                    lr=self.cfg.optimizer_params["learning_rate"],
                    update_epoch=lr_params.get(
                        "update_epoch",
                        self.cfg.end_epoch - self.cfg.begin_epoch - 1
                    )
                )
                lr_update_params.update(lr_params)

                trainer = module.Module.get_trainer(
                    net, optimizer=self.cfg.optimizer,
                    optimizer_params=self.cfg.optimizer_params,
                    lr_params=lr_update_params,
                    select=self.cfg.train_select,
                    logger=self.cfg.logger
                )

            try:
                train_time = toolbox["monitor"]["progress"].iteration_time
            except (KeyError, TypeError):
                train_time = None

            if (epoch - 1) % eval_every_n_epoch == 0 or epoch == end_epoch - 1:
                # # todo 定义每一轮结束后的模型评估方法
                evaluation_result = self.eval(
                    net, test_data, ctx=ctx
                )

                evaluation_formatter = formatter.get('evaluation', Formatter()) if formatter else Formatter()

                print(
                    evaluation_formatter(
                        iteration=epoch,
                        train_time=train_time,
                        loss_name_value=loss_values,
                        eval_name_value=evaluation_result,
                        extra_info=None,
                        dump=True,
                        keep="msg",
                    )
                )

            # todo 定义模型保存方案
            if save_model:
                if epoch % kwargs.get('save_epoch', 1) == 0:
                    self.save_params(
                        self.epoch_params_filepath(epoch), net
                    )

    def batch_loop(
            self,
            net, epoch, batch_size,
            train_data,
            trainer,
            loss_function,
            ctx=mx.cpu(),
            toolbox=None,
    ):
        """
        The true body of batch loop

        Parameters
        ----------
        net: HybridBlock
            The network which has been initialized
            or loaded from the existed model
        epoch: int
            Current Epoch
        batch_size: int
            The size of each batch
        train_data: Iterable
            The data used for this train procession,
            NOTICE: should have been divided to batches
        trainer:
            The trainer used to update the parameters of the net
        loss_function: dict of function
            The functions to compute the loss for the procession
            of back propagation
        ctx: Context or list of Context
            Defaults to ``mx.cpu()``.
        toolbox: dict
            Default to ``None``

        Returns
        -------
        loss_values: dict
            Recorded the loss values computed by the loss_function
        """
        loss_monitor = None
        progress_monitor = None

        if toolbox is not None:
            monitor = toolbox.get("monitor")
            if monitor is not None:
                loss_monitor = monitor.get("loss")
                progress_monitor = monitor.get("progress")
            if progress_monitor is None:
                def progress_monitor(x, y):
                    return tqdm(x, "[%s]" % y)

        for i, batch_data in progress_monitor(enumerate(train_data), epoch):
            self.fit_f(
                net=net, batch_size=batch_size, batch_data=batch_data,
                trainer=trainer,
                loss_function=loss_function,
                loss_monitor=loss_monitor,
                ctx=ctx,
            )

        if loss_monitor is not None:
            loss_values = {
                name: loss for name, loss in loss_monitor.items()
            }
            # optional
            loss_monitor.reset()
        else:
            loss_values = {}
        return i, loss_values

    @staticmethod
    def eval(net, test_data, ctx=mx.cpu()):
        """
        在这里定义数据评估方法

        Parameters
        ----------
        net
        test_data
        ctx

        Returns
        -------
        metrics: dict

        """
        return eval_f(net, test_data, ctx)



# coding: utf-8
# 2020/4/17 @ tongshiwei

import asyncio
from concurrent.futures import ThreadPoolExecutor as StdTPE
from multiprocessing.pool import Pool

__all__ = ["concurrent_pool", "ThreadPool", "ProcessPool", "CoroutinePool"]

SERIAL_LEVEL = 0
PROCESS_LEVEL = 1
THREAD_LEVEL = 2
COROUTINE = 3

_LEVEL = dict(
    d=SERIAL_LEVEL,
    debug=SERIAL_LEVEL,
    n=SERIAL_LEVEL,
    s=SERIAL_LEVEL,
    serial=SERIAL_LEVEL,
    t=THREAD_LEVEL,
    thread=THREAD_LEVEL,
    threading=THREAD_LEVEL,
    p=PROCESS_LEVEL,
    process=PROCESS_LEVEL,
    processing=PROCESS_LEVEL,
    c=COROUTINE,
    coro=COROUTINE,
    coroutine=COROUTINE,
)


class ThreadPool(StdTPE):
    def __init__(self, max_workers=None, thread_name_prefix='', ret: list = None):
        """
        Initializes a new ThreadPoolExecutor instance.

        Parameters
        ----------
        max_workers:
            The maximum number of threads that can be used to execute the given calls.
        thread_name_prefix:
            An optional name prefix to give our threads.
        """
        super(ThreadPool, self).__init__(max_workers=max_workers, thread_name_prefix=thread_name_prefix)
        self._result = []
        self.ret = ret

    def submit(self, fn, *args, **kwargs):
        _result = super(ThreadPool, self).submit(fn, *args, **kwargs)
        self._result.append(_result)
        return _result

    def __exit__(self, exc_type, exc_val, exc_tb):
        for r in self._result:
            if r.exception() is not None:
                raise r.exception()
            elif self.ret is not None:
                self.ret.append(r.result())

        super(ThreadPool, self).__exit__(exc_type, exc_val, exc_tb)


class ProcessPool(Pool):
    def __init__(self, processes=None, *args, ret: list = None, **kwargs):
        super(ProcessPool, self).__init__(processes, *args, **kwargs)
        self._result = []
        self.ret = ret

    def submit(self, func, *args, **kwargs):
        _result = self.apply_async(func, args, kwargs)
        self._result.append(_result)
        return _result

    def __exit__(self, exc_type, exc_val, exc_tb):
        for r in self._result:
            _r = r.get()
            if self.ret is not None:
                self.ret.append(_r)
        super(ProcessPool, self).__exit__(exc_type, exc_val, exc_tb)


class CoroutinePool(object):
    def __init__(self, *args, ret: list = None, **kwargs):
        self.pools = []
        self.ret = ret
        self._e = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        loop = asyncio.get_event_loop()
        self._e = asyncio.gather(*[func(*arg, **kwd) for func, arg, kwd in self.pools])
        loop.run_until_complete(self._e)
        loop.close()
        if self.ret is not None:
            for r in self._e.result():
                self.ret.append(r)

    def submit(self, func, *args, **kwargs):
        return self.pools.append([func, args, kwargs])


class SerialPool(object):
    def __init__(self, *args, ret=None, **kwargs):
        self.ret = ret

    def __enter__(self):
        return self

    def submit(self, fn, *args, **kwargs):
        r = fn(*args, **kwargs)
        if self.ret is not None:
            self.ret.append(r)
        return r

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


LEVEL_CLASS = {
    SERIAL_LEVEL: SerialPool,
    PROCESS_LEVEL: ProcessPool,
    THREAD_LEVEL: ThreadPool,
    COROUTINE: CoroutinePool,
}


def concurrent_pool(level: str, pool_size: int = None, ret: list = None):
    """
    Simple api for start completely independent concurrent programs:

    * thread
    * process
    * coroutine

    Examples
    --------
    .. code-block :: python

        def pseudo(idx):
            return idx

    .. code-block :: python

        ret = []
        with concurrent_pool("p", ret=ret) as e:  # or concurrent_pool("t", ret=ret)
             for i in range(4):
                e.submit(pseudo, i)
        print(ret)

    ``[0, 1, 2, 3]``


    """
    _level = _LEVEL[level]

    return LEVEL_CLASS[_level](pool_size, ret=ret)



# coding: utf-8
# created by tongshiwei on 18-2-5
import math
import sys
import warnings
from collections import OrderedDict
from contextlib import contextmanager

from longling import Timer
from longling.lib import ProgressMonitor, IterableMIcing
from longling.lib.stream import flush_print

try:
    NAN = math.nan
except (AttributeError, ImportError):  # pragma: no cover
    NAN = float('nan')

__all__ = ["ConsoleProgressMonitor", "ConsoleProgressMonitorPlayer"]


class ConsoleProgressMonitorPlayer(object):
    def __init__(self, indexes: (dict, OrderedDict), values: (dict, None) = None,
                 total=NAN, silent=False, timer=None, stream=sys.stdout, *args, **kwargs):
        """

        Parameters
        ----------
        indexes: dict
            {prefix: list of names}
        values: dict
            {prefix: list of functions}
        total
        silent
        timer
        stream
        args
        kwargs
        """

        if values is not None:
            assert type(indexes) == type(values)

        if isinstance(indexes, dict):
            indexes = OrderedDict(indexes)

        if values is not None:
            for prefix, names in indexes.items():
                for name in names:
                    _ = values[prefix][name]

        self.indexes = indexes

        self.n = None
        self.total = total
        self._total = None

        self.stream = stream
        self.timer = timer if timer is not None else Timer()

        index_header = ""
        arguments = []

        for prefix, _indexes in self.indexes.items():
            __indexes = ["%s-%s" % (prefix, _index) for _index in _indexes]
            arguments += __indexes
            index_header += (" " * 2).join(
                ["{:>%s}" % max(len(index), 15) for index in __indexes]
            )

        self.output_formatter = self.info_header + " " * 2 + index_header + " " * 2 + self.progress_header_fmt

        self.index_prefix = self.output_formatter.format(*self.headers, *arguments, self.progress_header)

        self.silent = silent
        self.values = values

    @property
    def progress_header_fmt(self):
        return "{:^30}"

    @property
    def progress_header(self):
        return "Progress"

    @property
    def progress(self):
        elapsed = self.timer.time()
        if not round(elapsed, 6) or not self.n:  # pragma: no cover
            return ""
        rate = self.n / elapsed
        elapsed = self.format_interval(int(elapsed))
        if self.total is not NAN:
            remaining = int((self.total - self.n) / rate)
            remaining = "<" + self.format_interval(remaining)
        else:
            remaining = ""

        return "[%s]" % (elapsed + remaining + ", {:.2f}it/s".format(rate))

    @staticmethod
    def format_interval(t):
        """
        Formats a number of seconds as a clock time, [H:]MM:SS

        Parameters
        ----------
        t  : int
            Number of seconds.

        Returns
        -------
        out  : str
            [H:]MM:SS

        Examples
        --------
        >>> ConsoleProgressMonitorPlayer.format_interval(360)
        '06:00'
        >>> ConsoleProgressMonitorPlayer.format_interval(100)
        '01:40'
        >>> ConsoleProgressMonitorPlayer.format_interval(36)
        '00:36'
        >>> ConsoleProgressMonitorPlayer.format_interval(3600)
        '1:00:00'
        >>> ConsoleProgressMonitorPlayer.format_interval(36000)
        '10:00:00'
        >>> ConsoleProgressMonitorPlayer.format_interval(360000)
        '100:00:00'
        >>> ConsoleProgressMonitorPlayer.format_interval(3600000)
        '1000:00:00'
        """
        mins, s = divmod(int(t), 60)
        h, m = divmod(mins, 60)
        if h:
            return '{0:d}:{1:02d}:{2:02d}'.format(h, m, s)
        else:
            return '{0:02d}:{1:02d}'.format(m, s)

    @property
    def info_header(self):
        return "{:>10}| {:>10}"

    @property
    def headers(self):
        return ["Iter", "Total-I"]

    def __call__(self, n, **kwargs):
        arguments = []

        for prefix, names in self.indexes.items():
            if prefix not in kwargs:
                ref = self.values[prefix]
            else:
                ref = kwargs[prefix]
            for name in names:
                arguments.append(ref[name])

        for prefix, name_value in kwargs.items():
            if prefix not in self.indexes:
                warnings.warn("detect unknown prefix: %s, all arguments will be ignored" % prefix)

        self.n = n

        res_str = self.output_format(*arguments)

        if not self.silent:
            flush_print(res_str, file=self.stream)

        self._total = n if self._total is None else max(n, self._total)

        return res_str

    def output_format(self, *arguments):
        return self.output_formatter.format(
            self.n, self.total, *arguments, self.progress
        )

    def start(self, *args, **kwargs):
        self.timer.start()

        res_str = self.index_prefix

        if not self.silent:
            print(res_str, file=self.stream)

        return res_str

    def end(self, n=None):
        if n is not None:
            self.total = n
        elif self._total is not None:
            self.total = self._total
        self(self.total)
        self.timer.end()
        if not self.silent:
            print("", file=self.stream)
        return ""

    @contextmanager
    def watching(self, *args, **kwargs):
        self.start(*args, **kwargs)
        yield
        self.end()

    @property
    def time(self):
        return self.timer.wall_time


class EBCPMP(ConsoleProgressMonitorPlayer):
    """
    Epoch-Batch Console Progress Monitor Player
    """

    def __init__(self, indexes: (dict, OrderedDict), values: (dict, None) = None,
                 total=NAN, silent=False, timer=None, stream=sys.stdout, total_epoch=NAN, *args, **kwargs):
        self.epoch = None
        self.total_epoch = total_epoch
        super(EBCPMP, self).__init__(
            indexes=indexes, values=values, total=total, silent=silent, timer=timer, stream=stream,
            *args, **kwargs
        )

    @property
    def info_header(self):
        return "{:>5}| {:>7}" + " " * 5 + "{:>10}" + " " * 2 + "{:>10}" + " " * 5

    @property
    def headers(self):
        return ["Epoch", "Total-E", "Batch", "Total-B"]

    def start(self, epoch, *args, **kwargs):
        self.epoch = epoch
        super(EBCPMP, self).start()

    def output_format(self, *arguments):
        return self.output_formatter.format(
            self.epoch, self.total_epoch, self.n, self.total, *arguments, self.progress
        )


class EpisodeCPMP(ConsoleProgressMonitorPlayer):
    @property
    def headers(self):
        return ["Episode", "Total-E"]


PLAYER_TYPE = {
    None: ConsoleProgressMonitorPlayer,
    "default": ConsoleProgressMonitorPlayer,
    "simple": ConsoleProgressMonitorPlayer,
    "epoch": EBCPMP,
    "episode": EpisodeCPMP,
}


class ConsoleProgressMonitor(ProgressMonitor):
    def __init__(self, indexes: (dict, OrderedDict), values: (dict, None) = None,
                 total=NAN, silent=False, player_type='default', *args, **kwargs):
        """

        Parameters
        ----------
        indexes: dict of list of str

        values: dict of dict of value

        total
        silent
        player_type
        args
        kwargs
        """
        super(ConsoleProgressMonitor, self).__init__(
            player=PLAYER_TYPE[player_type](
                indexes=indexes,
                values=values,
                total=total,
                silent=silent,
                *args,
                **kwargs
            )
        )
        self.silent = silent

    def __call__(self, iterator, *args, **kwargs):
        if self.silent:
            return iterator
        else:
            try:
                total = len(iterator)
                self.player.total = total
            except TypeError:
                pass
            self.player.start(*args, **kwargs)
            return IterableMIcing(iterator, self.player, self.player.end)

    @property
    def iteration_time(self):
        return self.player.time


if __name__ == '__main__':
    # def iter_fn(num):
    #     for i in range(num):
    #         yield i
    #
    #
    # import time
    # from longling import print_time
    #
    # values = {
    #     "Loss": {"l2": 0}
    # }
    # cp = ConsoleProgressMonitor({"Loss": ["l2"]}, values=values, player_type="epoch", total_epoch=2)
    #
    # with print_time():
    #     for e in range(2):
    #         for i in cp(iter_fn(100), e + 1):
    #             values["Loss"]["l2"] = i
    #             time.sleep(0.05)
    #
    #         print(cp.iteration_time)
    _values = {
        "Loss": {"l2": 0}
    }
    player = ConsoleProgressMonitorPlayer({"Loss": ["l2"]}, values=_values)
    with player.watching():
        for i in range(100):
            player(i, Loss={"l2": 1})
    print(player.time)



# coding: utf-8
# create by tongshiwei on 2020-11-12
import mxnet as mx
from mxnet import autograd

from .ctx import split_and_load

__all__ = ["fit_wrapper"]


def fit_wrapper(_fit_f):
    def fit_f(net, batch_data, loss_function, trainer,
              batch_size=None, loss_monitor=None,
              ctx=mx.cpu()):
        """
        Defined how each step of batch train goes

        Parameters
        ----------
        net: HybridBlock
            The network which has been initialized
            or loaded from the existed model
        batch_size: int
                The size of each batch
        batch_data: Iterable
            The batch data for train
        trainer:
            The trainer used to update the parameters of the net
        loss_function: dict of function
            The function to compute the loss for the procession
            of back propagation
        loss_monitor: LossMonitor
            Default to ``None``
        ctx: Context or list of Context
            Defaults to ``mx.cpu()``.

        Returns
        -------

        """
        batch_size = batch_size if batch_size is not None else len(batch_data[0])

        ctx_data = split_and_load(
            ctx, *batch_data,
            even_split=False
        )

        with autograd.record():
            for _data in ctx_data:
                bp_loss = _fit_f(
                    net, _data, loss_function, loss_monitor
                )
                assert bp_loss is not None
                bp_loss.backward()

        trainer.step(batch_size)

    return fit_f



# coding: utf-8
# create by tongshiwei on 2020-11-12

import functools

from longling.ML.MxnetHelper.glue import MetaModule, MetaModel, fit_f as meta_fit_f

__all__ = ["get_mx_module", "get_mx_model", "to_fit_f"]


def to_fit_f(fit_step_func):
    @functools.wraps(meta_fit_f)
    def fit_f(*args, **kwargs):
        return meta_fit_f(*args, fit_step_func=fit_step_func, **kwargs)

    return fit_f


def get_mx_module(
        module_cls: type(MetaModule) = MetaModule,
        net_gen_func=None,
        fit_func=None,
        net_init_func=None,
        eval_func=None
) -> type(MetaModule):
    """

    Parameters
    ----------
    module_cls
    net_gen_func
    fit_func
    net_init_func
    eval_func

    Returns
    -------

    """
    net_gen_func = module_cls.sym_gen if net_gen_func is None else net_gen_func
    fit_func = module_cls.fit_f if fit_func is None else fit_func
    net_init_func = module_cls.net_initialize if net_init_func is None else net_init_func
    eval_func = module_cls.eval if eval_func is None else eval_func

    class MetaModule(module_cls):
        @functools.wraps(net_gen_func)
        def sym_gen(self, *args, **kwargs):
            return net_gen_func(*args, **kwargs)

        @functools.wraps(fit_func)
        def fit_f(self, *args, **kwargs):
            return fit_func(*args, **kwargs)

        @functools.wraps(net_init_func)
        def net_initialize(self, *args, **kwargs):
            return net_init_func(*args, **kwargs)

        @staticmethod
        @functools.wraps(eval_func)
        def eval(*args, **kwargs):
            return eval_func(*args, **kwargs)

    return MetaModule


def get_mx_model(
        model_cls: type(MetaModel) = MetaModel,
        module_cls=None,
        configuration_cls=None,
        configuration_parser_cls=None,
) -> type(MetaModel):
    module_cls = MetaModel.get_module_cls() if module_cls is not None else model_cls
    configuration_cls = MetaModel.get_configuration_cls() if configuration_cls is not None else configuration_cls
    configuration_parser_cls = MetaModel.get_configuration_parser_cls(
    ) if configuration_parser_cls is not None else configuration_parser_cls

    class Model(model_cls):
        @staticmethod
        def get_module_cls():
            return module_cls

        @staticmethod
        def get_configuration_cls():
            return configuration_cls

        @staticmethod
        def get_configuration_parser_cls():
            return configuration_parser_cls

    return Model



# coding: utf-8
# create by tongshiwei on 2019/4/7

from .ctx import split_and_load
from .fit import fit_wrapper
from .init import load_net, net_initialize
from .lr_scheduler import get_lr_scheduler
from longling.ML.DL import get_optimizer_cfg
from .persistence import save_params
from .toolbox import get_default_toolbox
from .trainer import get_trainer
from .loss import as_tmt_mx_loss, loss_dict2tmt_mx_loss



# coding: utf-8
# 2019/9/20 @ tongshiwei

from .py_proj import py_proj
from .docs_proj import docs_proj
from .web_proj import web_proj

project_types = {
    "python": py_proj,
    "docs": docs_proj,
    "web": web_proj,
}



# coding: utf-8
# 2020/3/23 @ tongshiwei

import urllib.request
from .utils import get_http_header, get_opener
from longling.spider.conf import logger

__all__ = ["get_html_code"]


def get_html_code(url):
    """
    get encoded html code from specified url
    """
    logger.debug("request %s" % url)
    req = urllib.request.Request(url=url, headers=get_http_header(), method='GET')
    opener = get_opener()
    ret = opener.open(req)
    assert ret.status == 200, "error when request %s, error code %s" % (url, ret.status)
    content = ret.read()
    assert content, "error when request %s, null content" % url
    try:
        content = content.decode('utf-8')
    except UnicodeDecodeError:  # pragma: no cover
        content = content.decode('utf-8', 'ignore')
    return content



# coding: utf-8
# create by tongshiwei on 2019-9-6

from longling.lib.parser import get_parsable_var, load_configuration, \
    var2exp, path_append, Configuration as Params, ConfigurationParser

__all__ = [
    "Configuration",
    "var2exp", "eval_var",
    "path_append", "ConfigurationParser",
]


def eval_var(variable):
    return eval(variable)


class Configuration(Params):
    @property
    def parsable_var(self):
        """
        获取可以进行命令行设定的参数

        Returns
        -------
        store_vars: dict
            可以进行命令行设定的参数
        """
        return get_parsable_var(
            self,
            parse_exclude={'logger'},
        )

    @classmethod
    def load_cfg(cls, cfg_path, **kwargs):
        with open(cfg_path) as f:
            params = load_configuration(
                f,
            )
        params.update(kwargs)
        return params

    @classmethod
    def load(cls, cfg_path, **kwargs):
        return cls(**cls.load_cfg(cfg_path, **kwargs))



# coding: utf-8
# 2020/4/14 @ tongshiwei
from contextlib import contextmanager

__all__ = ["simulate_stdin"]


@contextmanager
def simulate_stdin(*inputs):
    """
    测试中模拟标准输入

    Parameters
    ----------
    inputs: list of str

    Examples
    --------
    >>> with simulate_stdin("12", "", "34"):
    ...     a = input()
    ...     b = input()
    ...     c = input()
    >>> a
    '12'
    >>> b
    ''
    >>> c
    '34'
    """
    from unittest.mock import patch
    from io import StringIO
    with patch('sys.stdin', StringIO("\n".join(inputs) + "\n")):
        yield



# coding: utf-8
# 2019/9/20 @ tongshiwei

__all__ = [
    "template_copy",
    "gitignore",
    "pytest", "coverage",
    "pysetup", "sphinx_conf", "makefile",
    "readthedocs", "travis", "nni",
    "dockerfile", "gitlab_ci", "chart"
]

import functools
import os
from collections import OrderedDict
from shutil import copyfile as _copyfile, rmtree, copytree as _copytree

from longling import wf_open, PATH_TYPE
from longling.Architecture.utils import binary_legal_input
from longling.lib.path import abs_current_dir, path_append
from longling.lib.regex import default_variable_replace as dvr
from longling.lib.utilog import config_logging, LogLevel
from longling.lib.yaml_helper import FoldedString, ordered_yaml_load, dump_folded_yaml

from . import config

logger = config_logging(logger="arch", console_log_level=LogLevel.INFO)

META = path_append(abs_current_dir(__file__), "meta_docs")
default_variable_replace = functools.partial(dvr, quotation="\'")


def copytree(src, dst, **kwargs):
    """
    Recursively copy a directory tree.

    Change OVERRIDE mode to specify the operation when dst existed.

    Examples
    --------
    .. code-block :: python

        # to change the OVERRIDE operation
        from longling.Architecture import config
        config.OVERRIDE = False  # skip existed dst
        config.OVERRIDE = None  # use console input to determine
        config.OVERRIDE = False  # override existed dst
    """
    if not override_check(dst):
        return
    elif os.path.exists(dst):
        rmtree(dst)
    _copytree(src, dst, **kwargs)


def copyfile(src, dst):
    """
    Copy data from src to dst.

    Change OVERRIDE mode to specify the operation when dst existed.

    Examples
    --------
    .. code-block :: python

        # to change the OVERRIDE operation
        from longling.Architecture import config
        config.OVERRIDE = False  # skip existed dst
        config.OVERRIDE = None  # use console input to determine
        config.OVERRIDE = False  # override existed dst
    """
    if override_check(dst):
        _copyfile(src, dst)


def override_check(path) -> bool:
    """
    Whether to override the specified file or directory

    not override:
     * config.OVERRIDE is False
     * config.OVERRIDE is None and user confirm not to override by console input

    Returns
    -------
    override: bool
        True to override while False not
    """
    if os.path.exists(path):
        if config.OVERRIDE is False:
            logger.error("%s exists, skipped" % path)
            return False
        elif config.OVERRIDE is None and not binary_legal_input("%s exists, override?" % path, _default="n"):
            logger.info("skip %s" % path)
            return False
        else:
            logger.info("%s exists, overrided" % path)
    return True


def template_copy(src: PATH_TYPE, tar: PATH_TYPE,
                  default_value: (str, dict, None) = "", quotation="\'", key_lower=True,
                  **variables):
    """
    Generate the tar file based on the template file where the variables will be replaced.
    Usually, the variable is specified like `$PROJECT` in the template file.


    Parameters
    ----------
    src: template file
    tar: target location
    default_value: the default value
    quotation: the quotation to wrap the variable value
    variables: the real variable values which are used to replace the variable in template file
    """

    if not override_check(tar):
        return

    with open(src) as f, wf_open(tar) as wf:
        for line in f:
            print(
                default_variable_replace(
                    line, default_value=default_value, quotation=quotation, key_lower=key_lower,
                    **variables
                ),
                end='', file=wf
            )


def gitignore(atype: str = "", tar_dir: PATH_TYPE = "./"):
    """
    cli alias: ``arch gitignore``


    Parameters
    ----------
    atype: the gitignore type, currently support `docs` and `python`
    tar_dir: target directory

    """
    src = path_append(META, "gitignore", "%s.gitignore" % atype)
    tar = path_append(tar_dir, ".gitignore")
    logger.info("gitignore: copy %s -> %s" % (src, tar))
    copyfile(src, tar)


def pytest(tar_dir: PATH_TYPE = "./"):
    """
    cli alias: ``arch pytest``

    Parameters
    ----------
    tar_dir

    Returns
    -------

    """
    src = path_append(META, "pytest.ini")
    tar = path_append(tar_dir, "pytest.ini")
    logger.info("pytest: copy %s -> %s" % (src, tar))
    copyfile(src, tar)


def coverage(tar_dir: PATH_TYPE = "./", **variables):
    """
    cli alias: ``arch coverage``

    Parameters
    ----------
    tar_dir:
    variables:
        These variables should be provided:

        * project

    """
    src = path_append(META, "setup.cfg.template")
    tar = path_append(tar_dir, "setup.cfg")
    logger.info("coverage: template %s -> %s" % (src, tar))
    template_copy(src, tar, quotation="", **variables)


def pysetup(tar_dir="./", **variables):
    """
    cli alias: ``arch pysetup``

    Parameters
    ----------
    tar_dir
    variables

    Returns
    -------

    """
    src = path_append(META, "setup.py.template")
    tar = path_append(tar_dir, "setup.py")
    logger.info("pysetup: template %s -> %s" % (src, tar))
    template_copy(src, tar, **variables)


def sphinx_conf(tar_dir="./", **variables):
    """
    cli alias: ``arch sphinx_conf``

    Parameters
    ----------
    tar_dir
    variables

    Returns
    -------

    """
    if variables["docs_style"] == "mxnet":
        src = path_append(META, "docs/mxnet/conf.py.template")
        tar = path_append(tar_dir, "conf.py")
        logger.info("sphinx_conf: template %s -> %s" % (src, tar))
        template_copy(src, tar, **variables)

        src = path_append(META, "docs/mxnet/.math.json")
        tar = path_append(tar_dir, ".math.json")
        logger.info("sphinx_conf: copy %s -> %s" % (src, tar))
        copyfile(src, tar)

        src = path_append(META, "docs/mxnet/requirements.txt")
        tar = path_append(tar_dir, "requirements.txt")
        logger.info("sphinx_conf: copy %s -> %s" % (src, tar))
        copyfile(src, tar)

        logger.warning(
            "\n%s\nmodify setup.py according to the components in %s\n%s\n" % ('*' * 60, tar, '*' * 60)
        )
    else:
        src = path_append(META, "docs/sphinx/requirements.txt")
        tar = path_append(tar_dir, "requirements.txt")
        logger.info("sphinx_conf: copy %s -> %s" % (src, tar))
        copyfile(src, tar)
        logger.warning(
            "\n%s\nmanually run 'sphinx-quickstart' in %s to create necessary components\n%s" % (
                '*' * 60, tar_dir, '*' * 60)
        )


def readthedocs(tar_dir="./"):
    """
    cli alias: ``arch readthedocs``

    Parameters
    ----------
    tar_dir

    Returns
    -------

    """
    src = path_append(META, "docs/.readthedocs.yml")
    tar = path_append(tar_dir, ".readthedocs.yml")
    logger.info("readthedocs: copy %s -> %s" % (src, tar))
    copyfile(src, tar)


def dockerfile(atype, tar_dir="./", **variables):
    """
    cli alias: ``arch dockerfile``

    Parameters
    ----------
    atype
    tar_dir
    variables

    Returns
    -------

    """
    src = path_append(META, "Dockerfile/%s.docker" % atype)
    tar = path_append(tar_dir, "Dockerfile")
    logger.info("Dockerfile:  %s -> %s" % (src, tar))
    template_copy(src, tar, quotation="", **variables)


def travis(tar_dir: PATH_TYPE = "./"):
    """
    cli alias: ``arch travis``

    Parameters
    ----------
    tar_dir

    Returns
    -------

    """
    src = path_append(META, ".travis.yml")
    tar = path_append(tar_dir, ".travis.yml")
    logger.info("travis: copy %s -> %s" % (src, tar))
    copyfile(src, tar)


def chart(tar_dir: PATH_TYPE = "./"):
    """
    cli alias: ``arch chart``

    Parameters
    ----------
    tar_dir: target directory

    """

    src_dir = path_append(META, "chart")
    tar_dir = path_append(tar_dir, "chart/")
    logger.info("chart: copy %s -> %s" % (src_dir, tar_dir))
    copytree(src_dir, tar_dir)


def helm_service(host="${KUBE_NAMESPACE}", image_repo="${CI_REGISTRY_IMAGE}", image_port=None, private=True,
                 name="$KUBE_NAMESPACE",
                 image_tag="latest", path_to_api=""):
    """

    Parameters
    ----------
    host
    image_repo
    image_port
    private
    name
    image_tag
    path_to_api

    Returns
    -------

    """
    src = path_append(META, "gitlab-ci/helm_install.gitlab-ci.yml.template")

    variables = {
        "NAME": name,
        "IMAGE_TAG": image_tag,
        "IMAGE_REPO": image_repo,
        "HOST": host,
        "PATH_TO_API": path_to_api,
    }

    with open(src, encoding="utf-8") as f:
        helm_install_commands = "".join(f.readlines()).strip() + "\n"
        helm_install_commands = default_variable_replace(helm_install_commands,
                                                         key_lower=False, quotation="", **variables)

    if image_port:
        helm_install_commands += "--set \"image.port=%s\"" % image_port + "\n"

    if private:
        helm_install_commands += "--set \"dockercfg=$(cat /root/.docker/config.json | base64 | tr -d '\\n')\"" + "\n"

    if not path_to_api:
        helm_install_commands += "--set \"ingress.annotations=null\"" + "\n"

    return FoldedString(helm_install_commands)


def _gitlab_ci(commands: dict, stage, image_name, private=True, on_stop=None, manual=False, only_master=False,
               deployment=True, registry_suffix="", version_in_path=True, **kwargs):
    name = "$KUBE_NAMESPACE"

    _commands = commands.get(stage, OrderedDict())
    _commands["image"] = image_name
    script = _commands.get("script", [])
    if script is None:  # pragma: no cover
        script = []
        _commands["script"] = script

    image_tag = "${API_VERSION}"
    ci_registry_image = "${CI_REGISTRY_IMAGE}"
    environment_name = stage

    export_version = r"export API_VERSION=$(cat chart/Chart.yaml | grep apiVersion | cut -d\  -f2)"

    if not only_master:
        image_tag = "${CI_COMMIT_REF_NAME}"
        environment_name += "/${CI_COMMIT_REF_NAME}"
    else:
        name += "-$API_VERSION"

    uninstall_heml = "helm uninstall %s || true" % name
    if deployment:
        docker_registry_image = "%s:%s" % ("${CI_REGISTRY_IMAGE}", image_tag)
        script[0] = dvr(script[0], key_lower=False,
                        **{"DOCKER_REGISTRY_IMAGE": docker_registry_image})
        if registry_suffix:
            ci_registry_image += registry_suffix
            script.insert(0, "export CI_REGISTRY_IMAGE=%s" % ci_registry_image)
        script.insert(0, export_version)

        script.append("helm dep build chart")
        script.append(uninstall_heml)
        if version_in_path:
            path_to_api = "${API_VERSION}(/|$)(.*)"
        else:
            path_to_api = ""
            kwargs.update({"host": "${KUBE_NAMESPACE}-%s" % "${API_VERSION}"})

        script.append(helm_service(name=name, private=private, path_to_api=path_to_api, image_tag=image_tag, **kwargs))

        environment = OrderedDict({"name": environment_name, "url": "https://$KUBE_NAMESPACE.env.bdaa.pro"})

        if on_stop is not None:
            environment["on_stop"] = "stop_%s" % stage
            if not manual:
                environment["auto_stop_in"] = "1 week"
            stop_commands = OrderedDict({
                "stage": "%s" % stage,
                "image": image_name,
                "script": uninstall_heml,
                "when": "manual",
                "variables": {"GIT_STRATEGY": "none"},
                "environment": OrderedDict({
                    "name": environment_name,
                    "action": "stop",
                }),
            })
        else:
            stop_commands = None

        _commands["environment"] = environment
        if only_master:
            _commands["only"] = OrderedDict({"refs": ["master"], "kubernetes": "active"})

        if manual:
            _commands["when"] = "manual"

        commands[stage] = _commands

        if stop_commands is not None:
            commands["stop_review"] = stop_commands


def gitlab_ci(private, stages: dict, atype: str = "", tar_dir: PATH_TYPE = "./", version_in_path=True):
    """
    cli alias: ``arch gitlab_ci``

    Parameters
    ----------
    private
    stages
    atype
    tar_dir
    version_in_path

    Returns
    -------

    """
    base_src = path_append(META, "gitlab-ci", ".gitlab-ci.yml")
    src = path_append(META, "gitlab-ci", "%s.gitlab-ci.yml" % atype)
    tar = path_append(tar_dir, ".gitlab-ci.yml")

    config_template = OrderedDict()

    with open(base_src) as f:
        config_template.update(ordered_yaml_load(f))

    with open(src) as f:
        config_template.update(ordered_yaml_load(f))

    logger.info("generate %s" % tar)

    with wf_open(tar) as wf:
        for _c in ["variables", "cache"]:
            if _c in config_template:
                print(dump_folded_yaml({_c: config_template[_c]}), file=wf)

        print(dump_folded_yaml({"stages": [stage for stage in stages.keys() if stage in config_template]}), file=wf)

        for stage, params in stages.items():
            if stage == "docs":
                params["registry_suffix"] = "/docs"
            elif stage in {"test", "build"}:
                params["deployment"] = False

            if stage not in config_template:
                logger.warning("%s is not listed in %s, skipped" % (stage, src))
                continue
            commands = {stage: config_template[stage]}
            _gitlab_ci(commands, stage, private=private, version_in_path=version_in_path, **params)
            print(dump_folded_yaml(commands), file=wf)

    if private:
        print("*" * 30)
        print("私有项目注意")
        print(
            "在项目的settings->repository->Deploy Tokens 添加一个name为gitlab-deploy-token、"
            "其他两项留空的具有read_registry权限的token"
        )
        print("*" * 30)


def makefile(tar_dir="./", **variables):
    """
    cli alias: ``arch makefile``

    Parameters
    ----------
    tar_dir
    variables

    Returns
    -------

    """
    src = path_append(META, "Makefile.template")
    tar = path_append(tar_dir, "Makefile")
    logger.info("makefile: template %s -> %s" % (src, tar))
    template_copy(src, tar, default_value=None, quotation='', **variables)


def nni(tar_dir="./"):
    """
    cli alias: ``arch nni`` and ``install nni``

    Parameters
    ----------
    tar_dir

    Returns
    -------

    """
    src_dir = path_append(META, "nni")
    for file in ["_config.yml", "_search_space.json"]:
        copyfile(path_append(src_dir, file), path_append(tar_dir, file))



# coding: utf-8
# 2020/4/13 @ tongshiwei

__all__ = ["dict_format", "pandas_format", "table_format", "series_format"]

from contextlib import contextmanager

from longling.lib.candylib import as_ordered_dict


def dict_format(data: dict, digits=6, col: int = None):
    """
    Examples
    --------
    >>> print(dict_format({"a": 123, "b": 3, "c": 4, "d": 5}))  # doctest: +NORMALIZE_WHITESPACE
    a: 123	b: 3	c: 4	d: 5
    >>> print(dict_format({"a": 123, "b": 3, "c": 4, "d": 5}, col=3))  # doctest: +NORMALIZE_WHITESPACE
    a: 123	b: 3	c: 4
    d: 5
    """
    if col is None:
        msg = []
        for name, value in data.items():
            if isinstance(value, float):
                _msg = "{}: {:.{digits}f}".format(name, value, digits=digits)
            else:
                _msg = "{}: {}".format(name, value)
            msg.append(_msg)
        msg = "\t".join([m for m in msg if m])
    else:
        msg = ""
        for i, (name, value) in enumerate(data.items()):
            if isinstance(value, float):
                _msg = "{}: {:.{digits}f}".format(name, value, digits=digits)
            else:
                _msg = "{}: {}".format(name, value)
            if (i + 1) % col == 0 and i != len(data) - 1:
                _msg += "\n"
            elif i != len(data) - 1:
                _msg += "\t"
            msg += _msg
    return msg


def pandas_format(data: (dict, list, tuple), columns: list = None, index: (list, str) = None, orient="index",
                  pd_kwargs: dict = None, max_rows=80, max_columns=80, **kwargs):
    """

    Parameters
    ----------
    data: dict, list, tuple, pd.DataFrame
    columns : list, default None
        Column labels to use when ``orient='index'``. Raises a ValueError
        if used with ``orient='columns'``.
    index : list of strings
        Optional display names matching the labels (same order).
    orient : {'columns', 'index'}, default 'columns'
            The "orientation" of the data. If the keys of the passed dict
            should be the columns of the resulting DataFrame, pass 'columns'
            (default). Otherwise if the keys should be rows, pass 'index'.
    pd_kwargs: dict
    max_rows: (int, None), default 80
    max_columns: (int, None), default 80

    Examples
    --------
    >>> print(pandas_format({"a": {"x": 1, "y": 2}, "b": {"x": 1.0, "y": 3}},  ["x", "y"]))
         x  y
    a  1.0  2
    b  1.0  3
    >>> print(pandas_format([[1.0, 2], [1.0, 3]],  ["x", "y"], index=["a", "b"]))
         x  y
    a  1.0  2
    b  1.0  3
    """

    import pandas as pd

    kwargs.update({
        "max_rows": max_rows,
        "max_columns": max_columns,
    })

    pd_kwargs = {} if pd_kwargs is None else pd_kwargs

    @contextmanager
    def pandas_session():
        for key, value in kwargs.items():
            pd.pandas.set_option(key, value)
        yield
        for key in kwargs:
            pd.pandas.reset_option(key)

    if isinstance(data, dict):
        data = as_ordered_dict(data, index)
        table = pd.DataFrame.from_dict(
            data, orient=orient, columns=columns, **pd_kwargs,
        )

    elif isinstance(data, (list, tuple, pd.DataFrame)):
        table = pd.DataFrame(data, index=index, columns=columns, **pd_kwargs)

    else:
        raise TypeError("cannot handle %s" % type(data))

    with pandas_session():
        return str(table)


series_format = dict_format
table_format = pandas_format
